#import modules.clean_data as cldt
#from ..modules import clean_data as clean
from text4gcn.modules import clean_data as clean
import time



class Metadata:
    '''
    Metadata class to define a decorator to save all the metadata related with the execution 
    of a function and save all to the database
    '''

    def metadata(key: str):
        '''
        function created to decorate another funcion
        :param key: name of the function (entry in the database)
        '''

        def metadata_decor(func: callable):
            '''
            decorator
            :param func: call function
            '''
            def wrapper(*args, **kwargs):
                '''
                wrapper
                '''
                args[0].logger.info(clean.create_title(key))

                initTime = time.time()
                args[0].logger.info(f"Start process: {initTime}")
                
                val = func(*args, **kwargs)

                endTime = time.time()
                totalTime = endTime-initTime

                args[0].logger.info("Elapsed time is %f seconds." % totalTime)
                args[0].logger.info()

                return val

            return wrapper

        return metadata_decor



from typing import List, Dict, MutableMapping, Tuple, Union
from collections import OrderedDict
from scipy.sparse import csr_matrix
#from ..modules import logger
from text4gcn.modules import logger
import pickle as pkl
import numpy as np


#from preprocessors.configs import PreProcessingConfigs
#from utils.file_ops import check_paths, create_dir
#from utils.common import check_data_set
#from utils.logger import PrintLog
#from time import time
#from math import ceil


# word -> word-vector, words are ordered with respect to vocabulary
WORD_VECTORS_TYPE = MutableMapping[str, np.ndarray]


class NodeFeatures():

    def __init__(self, logger: logger.PrintLog):
        self.logger = logger

    def extract_doc_labels(self, ds_corpus_meta_file: str) -> List[str]:
        with open(ds_corpus_meta_file) as ds_corpus_meta:
            doc_labels = list(OrderedDict.fromkeys(
                doc_meta.split()[2] for doc_meta in ds_corpus_meta))
        return doc_labels

    def compute_x(self, docs_of_words: List[List[str]], tr_size: int, emb_dim: int, w_vectors: WORD_VECTORS_TYPE) -> csr_matrix:
        """ x: feature vectors of training docs, no initial features """
        data_x = []
        for i in range(tr_size):
            doc_vec = np.zeros(emb_dim, dtype=float)  # Initialize
            words = docs_of_words[i]
            for word in words:
                if word in w_vectors:
                    doc_vec += w_vectors[word]

            mean_doc_vec = (doc_vec / len(words)).tolist()
            data_x.extend(mean_doc_vec)

        row_indexes = np.array(
            [[i] * emb_dim for i in range(tr_size)]).flatten().tolist()
        col_indexes = list(range(emb_dim)) * tr_size
        return csr_matrix((data_x, (row_indexes, col_indexes)), shape=(tr_size, emb_dim))

    def compute_y(self, doc_meta_list: List[str], train_size: int, doc_labels: List[str]) -> np.ndarray:
        y = []
        for i in range(train_size):
            doc_meta = doc_meta_list[i]
            one_hot_encoded_label = [0] * len(doc_labels)

            label = doc_meta.split('\t')[2]
            label_index = doc_labels.index(label)
            one_hot_encoded_label[label_index] = 1
            y.append(one_hot_encoded_label)
        return np.array(y)

    def compute_tx(self, docs_of_words: List[List[str]], test_size: int, real_train_size: int, word_emb_dim: int,
                   w_vectors: WORD_VECTORS_TYPE) -> csr_matrix:
        """ 
        tx: feature vectors of test docs, no initial features 
        """
        data_tx = []
        for i in range(test_size):
            doc_vec = np.zeros(word_emb_dim, dtype=float)  # Initialize
            words = docs_of_words[i + real_train_size]
            for word in words:
                if word in w_vectors:
                    doc_vec += w_vectors[word]

            mean_doc_vec = (doc_vec / len(words)).tolist()
            data_tx.extend(mean_doc_vec)

        row_indexes = np.array(
            [[i] * word_emb_dim for i in range(test_size)]).flatten().tolist()
        col_indexes = list(range(word_emb_dim)) * test_size
        return csr_matrix((data_tx, (row_indexes, col_indexes)), shape=(test_size, word_emb_dim))

    def compute_ty(self, doc_meta_list: List[str], test_size: int, real_train_size: int, doc_labels: List[str]) -> np.ndarray:
        ty = []
        for i in range(test_size):
            doc_meta = doc_meta_list[i + real_train_size]
            one_hot_encoded_label = [0] * len(doc_labels)

            label = doc_meta.split('\t')[2]
            label_index = doc_labels.index(label)
            one_hot_encoded_label[label_index] = 1
            ty.append(one_hot_encoded_label)
        return np.array(ty)

    def compute_allx(self, docs_of_words: List[List[str]], real_train_size: int, vocab: List[str],
                     word_vectors: WORD_VECTORS_TYPE, emb_dim: int) -> csr_matrix:
        """
        allx: A superset of x, the feature vectors of both labeled and words (unlabeled training instances)
        """
        word_vectors_arr = self.extract_word_vectors_arr(
            word_vectors, vocab, emb_dim=emb_dim)
        data_allx = []
        row_size = real_train_size + len(vocab)

        for i in range(real_train_size):
            doc_vec = np.zeros(emb_dim, dtype=float)  # Initialize
            words = docs_of_words[i]
            for word in words:
                if word in word_vectors:
                    doc_vec += word_vectors[word]

            mean_doc_vec = (doc_vec / len(words)).tolist()
            data_allx.extend(mean_doc_vec)

        data_allx.extend(word_vectors_arr.flatten())
        data_allx = np.array(data_allx)
        row_indexes = np.array(
            [[i] * emb_dim for i in range(row_size)]).flatten()
        col_indexes = np.array(list(range(emb_dim)) * row_size)

        return csr_matrix((data_allx, (row_indexes, col_indexes)), shape=(row_size, emb_dim))

    def compute_ally(self, doc_meta_list: List[str], real_train_size: int, doc_labels: List[str], vocab_size: int) -> np.ndarray:
        ally = []
        for doc_meta in doc_meta_list[:real_train_size]:
            label = doc_meta.split('\t')[2]
            one_hot_encoded_label = [0] * len(doc_labels)
            label_index = doc_labels.index(label)
            one_hot_encoded_label[label_index] = 1
            ally.append(one_hot_encoded_label)

        zero_filled_one_hot_for_words = [
            np.zeros(len(doc_labels), dtype=int)] * vocab_size
        ally.extend(zero_filled_one_hot_for_words)
        return np.array(ally)

    def load_word_to_word_vectors(self, path: str) -> Tuple[WORD_VECTORS_TYPE, int]:
        word_vectors_as_list = pkl.load(file=open(path, 'rb'))
        word_vectors = OrderedDict((word, np.array(vec_lst))
                                   for word, vec_lst in word_vectors_as_list.items())
        word_embedding_dimension = len(next(iter(word_vectors.values())))
        return word_vectors, word_embedding_dimension

    def extract_word_vectors_arr(self, w_vectors: WORD_VECTORS_TYPE, vocab: List[str], emb_dim: int) -> np.ndarray:
        np.random.seed(0)  # For reproducibility
        word_vectors_arr = np.random.uniform(-0.01,
                                             0.01, (len(vocab), emb_dim))
        if len(w_vectors) != 0:
            for i, word in enumerate(vocab):
                if word in w_vectors:
                    word_vectors_arr[i] = w_vectors[word]
        return word_vectors_arr

    def dump_node_features(self, directory: str, ds: str, node_features_dict: Dict[str, Union[np.ndarray, csr_matrix]]):
        for name, node_feature_matrix in node_features_dict.items():
            with open(f"{directory}/ind.{ds}.{name}", 'wb') as file:
                pkl.dump(node_feature_matrix, file)



from collections import Counter
from typing import List, Set
import re


def config_nltk():
    # temporary_nltk_folder = 'venv/nltk_data/'
    # if not os.path.exists(temporary_nltk_folder):
    #     from nltk import download
    #     download(info_or_id='stopwords', download_dir=temporary_nltk_folder)
    #     download(info_or_id='wordnet', download_dir=temporary_nltk_folder)
    from nltk import download

    #from nltk.corpus import wordnet
    #from nltk.corpus import stopwords
    download(info_or_id='stopwords')
    download(info_or_id='wordnet')
    download(info_or_id='omw-1.4')


def create_title(title):
    txt = f" {title} "
    return txt.center(100, "=")


def clean_str(a_str: str) -> str:
    """
    Tokenizing/string cleaning for all data-sets except for SST.
    Original taken from https://github.com/yoonkim/CNN_sentence/blob/master/process_data.py
    """
    string = re.sub(r"[^A-Za-z0-9(),!?\'`]", " ", a_str)
    string = re.sub(r"\'s", " \'s", string)
    string = re.sub(r"\'ve", " \'ve", string)
    string = re.sub(r"n\'t", " n\'t", string)
    string = re.sub(r"\'re", " \'re", string)
    string = re.sub(r"\'d", " \'d", string)
    string = re.sub(r"\'ll", " \'ll", string)
    string = re.sub(r",", " , ", string)
    string = re.sub(r"!", " ! ", string)
    string = re.sub(r"\(", r" \( ", string)
    string = re.sub(r"\)", r" \) ", string)
    string = re.sub(r"\?", r" \? ", string)
    string = re.sub(r"\s{2,}", " ", string)
    return string.strip().lower()


def retrieve_stop_words(language: str = 'english') -> Set[str]:
    #temporary_nltk_folder = 'venv/nltk_data/'
    from nltk.corpus import stopwords

    return set(stopwords.words(language))


def extract_word_counts(docs_of_words: List[List[str]]) -> Counter:
    """Extract word counts"""
    word_counts = Counter()
    for words in docs_of_words:
        word_counts.update(words)
    return word_counts


def remove_stop_words(lines_of_words: List[List[str]], stop_words: Set[str]) -> List[List[str]]:
    """ If a word is in stop-words, then remove it"""
    return [[word for word in line if word not in stop_words] for line in lines_of_words]


def remove_rare_words(lines_of_words: List[List[str]], word_counts: Counter, rare_count: int) -> List[List[str]]:
    """ If a word is rare, then remove it"""
    return [[word for word in line if word_counts[word] >= rare_count] for line in lines_of_words]


def glue_lines(lines_of_words: List[List[str]], glue_str: str, with_strip: bool) -> List[str]:
    if with_strip:
        return [glue_str.join(lines).strip() for lines in lines_of_words]
    else:
        return [glue_str.join(lines) for lines in lines_of_words]



from stanfordcorenlp import StanfordCoreNLP
from text4gcn.modules import file_ops as flop
from text4gcn.modules import logger as logger
from text4gcn.modules.word_processor import *
from text4gcn.modules.logger import Process
from scipy.sparse import csr_matrix
from nltk.corpus import stopwords
from math import log
import pickle


class DependencyParsingAdjacency():

    def __init__(self, dataset_name, dataset_path, core_nlp_path):
        self.dataset_name = dataset_name
        self.dataset_path = dataset_path
        self.core_nlp_path = core_nlp_path
        self.logger = logger.PrintLog()
        self.flop = flop.FileOps(logger=self.logger)

    def build_relation_pair(self, docs_of_words):
        stop_words = set(stopwords.words('english'))
        nlp = StanfordCoreNLP(self.core_nlp_path, lang='en')

        errors = 0
        rela_pair_count_str = {}
        for docs_of_word in docs_of_words:
            words = docs_of_word
            sentence = ' '.join(words)

            try:
                res = nlp.dependency_parse(sentence)
                tokenized = nlp.word_tokenize(sentence)
            except Exception as e:
                #print(f'{sentence} = {e}')
                errors += 1
                res = []

            for pair in list(res):
                #pair=pair.split(", ")
                if pair[0] == 'ROOT' or pair[1] == 'ROOT':
                    continue
                if pair[0] == pair[1]:
                    continue
                # if pair[0] in string.punctuation or pair[1] in string.punctuation:
                #    continue
                if pair[0] in stop_words or pair[1] in stop_words:
                    continue

                word_pair_str = f'{tokenized[pair[2]-1]},{tokenized[pair[1]-1]}'
                if word_pair_str in rela_pair_count_str:
                    rela_pair_count_str[word_pair_str] += 1
                else:
                    rela_pair_count_str[word_pair_str] = 1
                # two orders
                word_pair_str = f'{tokenized[pair[1]-1]},{tokenized[pair[2]-1]}'
                if word_pair_str in rela_pair_count_str:
                    rela_pair_count_str[word_pair_str] += 1
                else:
                    rela_pair_count_str[word_pair_str] = 1
        print(f'[INFO] Error in Dependency Parse: {errors}')
        nlp.close()
        return rela_pair_count_str

    def compute_weights_with_PMI(self, docs_of_words, rela_pair_count_str, word_window_freq, train_size, word_id_map, min_count1, max_count1, count_mean1, count_std1, row, col):
        weight = []
        errors = 0
        num_window = len(docs_of_words)
        for key, count in rela_pair_count_str.items():
            try:
                temp = key.split(',')
                i = temp[0]
                j = temp[1]

                if i in word_window_freq and j in word_window_freq:
                    word_freq_i = word_window_freq[i]
                    word_freq_j = word_window_freq[j]
                    pmi = log(
                        1.0
                        * count
                        / num_window
                        / (1.0 * word_freq_i * word_freq_j / num_window ** 2)
                    )
                    if pmi <= 0:
                        continue

                    row.append(train_size + word_id_map[i])
                    col.append(train_size + word_id_map[j])
                    if key in rela_pair_count_str:
                        wei = (rela_pair_count_str[key] -
                               min_count1) / (max_count1 - min_count1)
                        wei = (
                            rela_pair_count_str[key]-count_mean1) / count_std1
                        weight.append(wei)
            except:
                errors += 1

        print(f'[INFO] Error in Compute Weights: {errors}')
        return weight

    def compute_weights(self, docs_of_words, rela_pair_count_str, word_window_freq, train_size, word_id_map, vocab_size, word_doc_freq, vocab):
        row = []
        col = []

        min_count1, max_count1, count_mean1, count_std1 = relation_pair_statitcs(
            rela_pair_count_str
        )

        # compute weights PMI
        weight = self.compute_weights_with_PMI(
            docs_of_words,
            rela_pair_count_str,
            word_window_freq,
            train_size,
            word_id_map,
            min_count1,
            max_count1,
            count_mean1,
            count_std1,
            row,
            col
        )

        # doc word frequency
        doc_word_freq = get_doc_word_freq(docs_of_words, word_id_map)

        weight_tfidf = get_weight_tfidf(
            docs_of_words,
            word_id_map,
            doc_word_freq,
            train_size,
            vocab_size,
            word_doc_freq,
            vocab,
            row,
            col
        )

        return row, col, weight, weight_tfidf

    @Process.log("EXTRACTED ADJACENCY MATRIX: Heterogenous doc-word adjacency matrix.")
    def build(self):

        corpus_path = f"{self.dataset_path}/{self.dataset_name}"
        ds_corpus = f'{corpus_path}.shuffled/{self.dataset_name}.txt'
        ds_corpus_vocabulary = f'{corpus_path}.shuffled/{self.dataset_name}.vocab'
        ds_corpus_train_idx = f'{corpus_path}.shuffled/{self.dataset_name}.train'
        ds_corpus_test_idx = f'{corpus_path}.shuffled/{self.dataset_name}.test'

        self.flop.create_dir(
            dir_path=f'{corpus_path}.adjacency',
            overwrite=False
        )

        docs_of_words = [line.split() for line in open(file=ds_corpus)]
        # Extract Vocabulary.
        vocab = open(ds_corpus_vocabulary).read().splitlines()
        # Real train-size, not adjusted.
        train_size = len(open(ds_corpus_train_idx).readlines())
        # Real test-size.
        test_size = len(open(ds_corpus_test_idx).readlines())

        word_doc_freq = define_word_doc_freq(docs_of_words)

        vocab_size = len(vocab)
        word_id_map = {}
        id_word_map = {}
        for i in range(vocab_size):
            word_id_map[vocab[i]] = i
            id_word_map[i] = vocab[i]

        word_window_freq = build_word_window_freq(docs_of_words)

        rela_pair_count_str = self.build_relation_pair(docs_of_words)

        row, col, weight, weight_tfidf = self.compute_weights(
            docs_of_words,
            rela_pair_count_str,
            word_window_freq,
            train_size,
            word_id_map,
            vocab_size,
            word_doc_freq,
            vocab
        )

        # =============================================================
        weight += weight_tfidf
        node_size = train_size + vocab_size + test_size

        self.logger.info(
            f"({len(weight)}, ({len(row)}, {len(col)})), shape=({node_size}, {node_size})")

        adj = csr_matrix((weight, (row, col)), shape=(node_size, node_size))
        # =============================================================

        # Dump Adjacency Matrix
        with open(f"{corpus_path}.adjacency/ind.dependency.{self.dataset_name}.adj", 'wb') as f:
            pickle.dump(adj, f)



import matplotlib.pyplot as plt
from tabulate import tabulate
# from ..models.utils import *
from text4gcn.models.utils import *
from sklearn import metrics
import scipy.sparse as sp
from time import time
import numpy as np
import torch as th


class TextGCNTrainer:
    def __init__(self, model, args, pre_data):
        self.args = args
        self.model = model
        self.device = args.device
        self.max_epoch = self.args.max_epoch
        self.nhid = self.args.nhid
        self.dataset = args.dataset
        self.predata = pre_data
        self.earlystopping = args.early_stopping
        self.log_dir = args.log_dir

    def fit(self):
        # Extract data
        self.prepare_data()
        # Load corpus, unpack values & convert to tensor
        self.convert_tensor()

        self.model = self.model(
            input_dim=self.features.shape[0], 
            output_dim=self.nhid, 
            support=self.support, 
            num_classes=self.y_train.shape[1]
        ).to(self.device)

        # Loss and optimizer
        self.criterion = th.nn.CrossEntropyLoss()
        self.optimizer = th.optim.Adam(
            self.model.parameters(), lr=self.args.lr)

        print("\nModel structure:\n\n", self.model, "\n")
        for name, param in self.model.named_parameters():
            print(f"Layer: {name} | Size: {param.size()}")

        self.train_model()

    def prepare_data(self):
        adj, features, y_train, y_val, y_test, train_mask, val_mask, test_mask, train_size, test_size = self.predata

        # featureless
        self.features = sp.identity(features.shape[0])
        # feature of nodes
        self.features = preprocess_features(self.features)
        # adjacency matrix
        self.adj = adj
        self.support = [preprocess_adj(adj)]

        self.train_mask = train_mask
        self.val_mask = val_mask
        self.test_mask = test_mask
        self.y_train = y_train
        self.y_val = y_val
        self.y_test = y_test
        self.train_size = train_size
        self.test_size = test_size

    def convert_tensor(self):

        print("\nData statistics\n")
        print(tabulate([
            ["Edges", "Classes", "Train samples", "Val samples", "Test samples"],
            [self.features.shape[0], self.y_train.shape[1], self.train_mask.sum(), self.val_mask.sum(), self.test_mask.sum()]],
            headers="firstrow"))

        self.features = th.from_numpy(self.features).float().to(self.device)
        self.y_train = th.from_numpy(self.y_train).float().to(self.device)
        self.y_val = th.from_numpy(self.y_val).float().to(self.device)
        self.y_test = th.from_numpy(self.y_test).float().to(self.device)
        self.train_mask = th.from_numpy(
            self.train_mask).float().to(self.device)
        self.val_mask = th.from_numpy(self.val_mask).float().to(self.device)
        self.test_mask = th.from_numpy(self.test_mask).float().to(self.device)
        self.tm_train_mask = th.transpose(th.unsqueeze(
            self.train_mask, 0), 1, 0).repeat(1, self.y_train.shape[1])
        self.support = [th.Tensor(self.support[i])
                        for i in range(len(self.support))]

    def evaluate_model(self, model, criterion, features, labels, mask):
        t_test = time()

        model.eval()
        with th.no_grad():
            logits = model(features)
            t_mask = th.from_numpy(np.array(mask * 1., dtype=np.float32))
            tm_mask = th.transpose(th.unsqueeze(
                t_mask, 0), 1, 0).repeat(1, labels.shape[1])
            loss = criterion(logits * tm_mask, th.max(labels, 1)[1])
            pred = th.max(logits, 1)[1]

            try:
                acc = ((pred == th.max(labels, 1)[1]).float(
                ) * t_mask).sum().item() / t_mask.sum().item()
            except ZeroDivisionError:
                acc = 0

        return loss.numpy(), acc, pred.numpy(), labels.numpy(), (time() - t_test)

    def show_metrics(self, test_labels, test_pred):
        """ Calucate evaluation metrics for precision, recall, and f1.

        Arguments
        ---------
            y_pred: ndarry, the predicted result list
            y_true: ndarray, the ground truth label list

        Returns
        -------
            precision: float, precision value
            recall: float, recall value
            f1: float, f1 measure value
        """
        print("\nTest Precision, Recall and F1-Score...")
        print(metrics.classification_report(test_labels, test_pred, digits=4))
        print("\nMacro average Test Precision, Recall and F1-Score...")
        print(metrics.precision_recall_fscore_support(
            test_labels, test_pred, average='macro'))
        print("\nMicro average Test Precision, Recall and F1-Score...")
        print(metrics.precision_recall_fscore_support(
            test_labels, test_pred, average='micro'))
        print("\nWeighted average Test Precision, Recall and F1-Score...")
        print(metrics.precision_recall_fscore_support(
            test_labels, test_pred, average='weighted'))
        print("\nConfusion Matrix...")
        print(metrics.confusion_matrix(test_labels, test_pred))
        print("\nF1-Score...")
        print(metrics.f1_score(test_labels, test_pred, average="macro"))
        print("\nPrecision-Score...")
        print(metrics.precision_score(test_labels, test_pred, average="macro"))
        print("\nRecall-Score...")
        print(metrics.recall_score(test_labels, test_pred, average="macro"))
        print("\nAccuracy-Score...")
        print(metrics.accuracy_score(test_labels, test_pred))

    def test(self, t1):
        # Testing
        test_loss, test_acc, pred, labels, test_duration = self.evaluate_model(
            self.model, self.criterion, self.features, self.y_test, self.test_mask)
        print("\nTest set results: loss= {:.5f}, accuracy= {:.5f}, time= {:.5f}".format(
            test_loss, test_acc, test_duration))

        test_pred = []
        test_labels = []
        for i in range(len(self.test_mask)):
            if self.test_mask[i]:
                test_pred.append(pred[i])
                test_labels.append(np.argmax(labels[i]))

        elapsed = time() - t1

        # print("\nTest Precision, Recall and F1-Score...")
        # print(metrics.classification_report(test_labels, test_pred, digits=4))
        # print("\nMacro average Test Precision, Recall and F1-Score...")
        # print(metrics.precision_recall_fscore_support(test_labels, test_pred, average='macro'))
        # print("\nMicro average Test Precision, Recall and F1-Score...")
        # print(metrics.precision_recall_fscore_support(test_labels, test_pred, average='micro'))
        # print("\nWeighted average Test Precision, Recall and F1-Score...")
        # print(metrics.precision_recall_fscore_support(test_labels, test_pred, average='weighted'))
        # print("\nConfusion Matrix...")
        # print(metrics.confusion_matrix(test_labels, test_pred))
        # print("\nF1-Score...")
        # print(metrics.f1_score(test_labels, test_pred, average="macro"))
        # print("\nPrecision-Score...")
        # print(metrics.precision_score(test_labels, test_pred, average="macro"))
        # print("\nRecall-Score...")
        # print(metrics.recall_score(test_labels, test_pred, average="macro"))
        # print("\nAccuracy-Score...")
        # print(metrics.accuracy_score(test_labels, test_pred))
        self.show_metrics(test_labels, test_pred)

        # doc and word embeddings
        tmp = self.model.layer1.embedding.numpy()
        word_embeddings = tmp[self.train_size: self.adj.shape[0] - self.test_size]
        train_doc_embeddings = tmp[:self.train_size]
        test_doc_embeddings = tmp[self.adj.shape[0] - self.test_size:]

        print('\nEmbeddings:')
        print(f'Word_embeddings:{len(word_embeddings)}')
        print(f'Train_doc_embeddings:{len(train_doc_embeddings)}')
        print(f'Test_doc_embeddings:{len(test_doc_embeddings)}')
        print("\nElapsed time is %f seconds." % elapsed)

    def saveModel(self):
        name = unique_name()
        #path = f"{self.log_dir}/NetModel.pth"
        path = f"{self.log_dir}/{unique_name()}.pth"
        th.save(self.model.state_dict(), path)

    def plot_accuracy(self, train_acc_hist, eval_acc_hist, train_loss_hist, eval_loss_hist):
        epochs = range(1, len(train_acc_hist) + 1)
        plt.figure(figsize=(16, 10))
        plt.plot(epochs, train_acc_hist, 'b', label='Training acc')
        plt.plot(epochs, eval_acc_hist, 'r', label='Validation acc')
        plt.title('Training and validation accuracy')
        plt.legend()
        plt.savefig(f'{self.log_dir}/metrics-acc.png', dpi=250)

        plt.figure()
        plt.figure(figsize=(16, 10))
        plt.plot(epochs, train_loss_hist, 'b', label='Training loss')
        plt.plot(epochs, eval_loss_hist, 'r', label='Validation loss')
        plt.title('Training and validation loss')
        plt.legend()
        plt.savefig(f'{self.log_dir}/metrics-loss.png', dpi=250)

    def train_model(self):

        t1 = time()

        th.manual_seed(self.args.seed)
        print(f"\n{'∎'*20} Torch Seed: {th.seed()}")  # "∎"

        train_loss_hist = []
        eval_loss_hist = []

        train_acc_hist = []
        eval_acc_hist = []

        val_losses = []

        # Train model
        for epoch in range(self.max_epoch):
            epoch_start_time = time()

            # Forward pass: Compute predicted y by passing x to the model
            logits = self.model(self.features)
            # Compute loss
            loss = self.criterion(
                logits * self.tm_train_mask, th.max(self.y_train, 1)[1])

            try:
                acc = ((th.max(logits, 1)[1] == th.max(
                    self.y_train, 1)[1]).float(
                ) * self.train_mask).sum().item() / self.train_mask.sum().item()
            except ZeroDivisionError:
                acc = 0

            # Backward and optimize
            # Zero gradients, perform a backward pass, and update the weights.
            self.optimizer.zero_grad()
            # for each parameter, calculate d(loss)/d(weight)
            loss.backward()
            # update weights, causes the optimizer to take a step based on the gradients of the parameters
            self.optimizer.step()

            # Validation
            val_loss, val_acc, pred, labels, duration = self.evaluate_model(
                self.model, self.criterion, self.features, self.y_val, self.val_mask)
            val_losses.append(val_loss)

            print("Epoch:{:04d}, train_loss={:.5f}, train_acc={:.5f}, val_loss={:.5f}, val_acc={:.5f}, time={:.5f}".format(
                epoch + 1, loss, acc, val_loss, val_acc, time() - epoch_start_time))

            if epoch > self.earlystopping and val_losses[-1] > np.mean(val_losses[-(self.earlystopping + 1):-1]):
                print(
                    f'Early stopping... Validation loss decreased ({val_loss:.6f}).  Saving model ...')
                self.saveModel()
                break

            train_loss_hist.append(float(loss))
            train_acc_hist.append(float(acc))
            eval_loss_hist.append(float(val_loss))
            eval_acc_hist.append(float(val_acc))

        print('\nCompleted training batch', epoch+1, 'Training Loss is: %.4f' %
              loss, 'Validation Loss is: %.4f' % val_loss, 'Accuracy is %.4f' % (acc))

        # epochs = range(1, len(train_acc_hist) + 1)
        # plt.figure(figsize=(16, 10))
        # plt.plot(epochs, train_acc_hist, 'b', label='Training acc')
        # plt.plot(epochs, eval_acc_hist, 'r', label='Validation acc')
        # plt.title('Training and validation accuracy')
        # plt.legend()
        # plt.savefig(f'{self.log_dir}/metrics-acc.png', dpi=250)

        # plt.figure()
        # plt.figure(figsize=(16, 10))
        # plt.plot(epochs, train_loss_hist, 'b', label='Training loss')
        # plt.plot(epochs, eval_loss_hist, 'r', label='Validation loss')
        # plt.title('Training and validation loss')
        # plt.legend()
        # plt.savefig(f'{self.log_dir}/metrics-loss.png', dpi=250)
        self.plot_accuracy(train_acc_hist, eval_acc_hist, train_loss_hist, eval_loss_hist)

        print("\nOptimization Finished!")

        self.test(t1)



from text4gcn.modules.textvec.word_features import train_word2vec
from text4gcn.modules import adjacency as adjcy
from text4gcn.modules import logger as logger
from text4gcn.modules import file_ops as flop
from scipy.sparse import csr_matrix
from nltk.corpus import stopwords
import pickle


class EmbeddingAdjacency():

    def __init__(self, dataset_name, dataset_path, training_regime, embedding_dimension, num_epochs):
        self.dataset_name = dataset_name
        self.dataset_path = dataset_path
        self.training_regime = training_regime
        self.embedding_dimension = embedding_dimension
        self.num_epochs = num_epochs
        self.logger = logger.PrintLog()
        self.flop = flop.FileOps(logger=self.logger)

    def build_relation_pair(self, docs_of_words):
        stop_words = set(stopwords.words('english'))

        word2vec_model = train_word2vec(
            save_dir=None,
            document_list=docs_of_words,
            num_epochs=20,
            embedding_dimension=300,
            training_regime=1,
        )

        errors = 0
        rela_pair_count_str = {}
        for docs_of_word in docs_of_words:
            docs_of_word_len = len(docs_of_word)
            for i in range(docs_of_word_len):
                if i+1 == docs_of_word_len:
                    continue

                if docs_of_word[i] == docs_of_word[i+1]:
                    continue

                if docs_of_word[i] in stop_words or docs_of_word[i+1] in stop_words:
                    continue

                cosine_similarity = word2vec_model.wv.similarity(
                    docs_of_word[i], docs_of_word[i+1])
                if cosine_similarity < 0.95:
                    errors += 1
                    continue

                word_pair_str = f'{docs_of_word[i]},{docs_of_word[i+1]}'
                if word_pair_str in rela_pair_count_str:
                    rela_pair_count_str[word_pair_str] += 1
                else:
                    rela_pair_count_str[word_pair_str] = 1
                # two orders
                word_pair_str = f'{docs_of_word[i+1]},{docs_of_word[i]}'
                if word_pair_str in rela_pair_count_str:
                    rela_pair_count_str[word_pair_str] += 1
                else:
                    rela_pair_count_str[word_pair_str] = 1

        print(f'[INFO] Words without similarity: {errors}')
        return rela_pair_count_str

    def compute_weights(self, docs_of_words, rela_pair_count_str, word_window_freq, train_size, word_id_map, vocab_size, word_doc_freq, vocab):
        row = []
        col = []

        min_count1, max_count1, count_mean1, count_std1 = adjcy.relation_pair_statitcs(
            rela_pair_count_str)

        # compute weights PMI
        weight = adjcy.compute_weights_with_PMI(
            docs_of_words, rela_pair_count_str, word_window_freq, train_size, word_id_map, min_count1, max_count1, count_mean1, count_std1, row, col)

        # doc word frequency
        doc_word_freq = adjcy.get_doc_word_freq(docs_of_words, word_id_map)

        weight_tfidf = adjcy.get_weight_tfidf(
            docs_of_words, word_id_map, doc_word_freq, train_size, vocab_size, word_doc_freq, vocab, row, col)

        return row, col, weight, weight_tfidf

    def build(self):

        corpus_path = f"{self.dataset_path}/{self.dataset_name}"
        ds_corpus = f'{corpus_path}.shuffled/{self.dataset_name}.txt'
        ds_corpus_vocabulary = f'{corpus_path}.shuffled/{self.dataset_name}.vocab'
        ds_corpus_train_idx = f'{corpus_path}.shuffled/{self.dataset_name}.train'
        ds_corpus_test_idx = f'{corpus_path}.shuffled/{self.dataset_name}.test'

        # checkers
        #check_data_set(data_set_name=ds_name, all_data_set_names=cfg.data_sets)
        #check_paths(ds_corpus, ds_corpus_vocabulary, ds_corpus_train_idx, ds_corpus_test_idx)

        #create_dir(dir_path=cfg.corpus_shuffled_adjacency_dir + "/semantic", overwrite=False)
        self.flop.create_dir(
            dir_path=f'{corpus_path}.adjacency', overwrite=False)

        docs_of_words = [line.split() for line in open(file=ds_corpus)]
        # Extract Vocabulary.
        vocab = open(ds_corpus_vocabulary).read().splitlines()
        # Real train-size, not adjusted.
        train_size = len(open(ds_corpus_train_idx).readlines())
        # Real test-size.
        test_size = len(open(ds_corpus_test_idx).readlines())

        word_doc_freq = adjcy.define_word_doc_freq(docs_of_words)

        vocab_size = len(vocab)
        word_id_map = {}
        id_word_map = {}
        for i in range(vocab_size):
            word_id_map[vocab[i]] = i
            id_word_map[i] = vocab[i]

        word_window_freq = adjcy.build_word_window_freq(docs_of_words)

        rela_pair_count_str = self.build_relation_pair(docs_of_words)

        row, col, weight, weight_tfidf = self.compute_weights(
            docs_of_words, rela_pair_count_str, word_window_freq, train_size, word_id_map, vocab_size, word_doc_freq, vocab)

        # =============================================================
        weight += weight_tfidf
        node_size = train_size + vocab_size + test_size

        self.logger.info(
            f"({len(weight)}, ({len(row)}, {len(col)})), shape=({node_size}, {node_size})")

        adj = csr_matrix((weight, (row, col)), shape=(node_size, node_size))
        # =============================================================
        # Dump Adjacency Matrix
        with open(f"{corpus_path}.adjacency/ind.embedding.{self.dataset_name}.adj", 'wb') as f:
            pickle.dump(adj, f)



from typing import Any, Iterable, List, Tuple
#from ..modules import logger
from text4gcn.modules import logger
from os.path import exists
from shutil import rmtree
from os import makedirs
import pickle as pkl


class FileOps():
    def __init__(self, logger: logger.PrintLog):
        self.logger = logger

    def create_dir(self, dir_path: str, overwrite: bool) -> None:
        if exists(dir_path):
            if overwrite:
                rmtree(dir_path)
                makedirs(dir_path)
            else:
                #print('[WARN] directory:%r already exists, not overwritten.' % dir_path)
                self.logger.warning(
                    'Directory:%r already exists, not overwritten.' % dir_path)
        else:
            makedirs(dir_path)

    def write_iterable_to_file(self, an_iterable: Iterable[Any], file_path: str, file_mode: str = 'w'):
        with open(file_path, file_mode) as f:
            f.writelines("%s\n" % item for item in an_iterable)

    def check_paths(self, *paths: str):
        """
        Check paths if they exist or not
        """

        for path in paths:
            if not exists(path):
                raise FileNotFoundError(
                    'Path: {path} is not found.'.format(path=path))

    def load_corpus_meta(self, corpus_meta_path: str) -> Tuple[List[str], List[str], List[str]]:
        all_doc_meta_list = [
            line.strip() for line in open(corpus_meta_path, 'r').readlines()]

        train_doc_meta_list = [
            doc_meta for doc_meta in all_doc_meta_list if doc_meta.split('\t')[
                1].endswith('train')]

        test_doc_meta_list = [
            doc_meta for doc_meta in all_doc_meta_list if doc_meta.split('\t')[
                1].endswith('test')]

        return all_doc_meta_list, train_doc_meta_list, test_doc_meta_list

    def write_picke(self, obj, file):
        pkl.dump(obj=obj, file=open(file, mode='wb'))


# def check_data_set(data_set_name: str, all_data_set_names: List[str]) -> None:
#     if data_set_name not in all_data_set_names:
#         raise AttributeError(
#             "Wrong data-set name, given:%r, however expected:%r" %
#             (data_set_name, all_data_set_names))



#from ..modules import clean_data as clean
from text4gcn.modules import clean_data as clean
import datetime
import time


INFO = "[INFO]"
ERROR = "[ERROR]"
WARNING = "[WARN]"


class PrintLog():

    def __init__(self):
        self.output = []

    def log(self, msg='', end='\n'):
        now = datetime.datetime.now()
        t = (
            (
                (
                    f'{str(now.year)}/{str(now.month)}/{str(now.day)} '
                    + str(now.hour).zfill(2)
                )
                + ':'
            )
            + str(now.minute).zfill(2)
            + ':'
        ) + str(now.second).zfill(2)

        lines = msg.split('\n') if isinstance(msg, str) else [msg]

        for line in lines:
            if line == lines[-1]:
                print(f'[{t}] {str(line)}', end=end)
            else:
                print(f'[{t}] {str(line)}')
            self.output.append(str(line)+end)

    def info(self, msg='', end='\n'):
        self.log(msg=f'{INFO} {msg}', end=end)

    def warning(self, msg='', end='\n'):
        self.log(msg=f'{WARNING} {msg}', end=end)

    def error(self, msg='', end='\n'):
        self.log(msg=f'{ERROR} {msg}', end=end)

    def log_history(self):
        return self.output


class Process:
    '''
    Metadata class to define a decorator to save all the metadata related with the execution 
    of a function and save all to the database
    '''

    def log(key: str):
        '''
        function created to decorate another funcion
        :param key: name of the function (entry in the database)
        '''

        def metadata_decor(func: callable):
            '''
            decorator
            :param func: call function
            '''
            def wrapper(*args, **kwargs):
                '''
                wrapper
                '''
                args[0].logger.info(clean.create_title(key))

                initTime = time.time()
                args[0].logger.info(f"Start process: {initTime}")

                val = func(*args, **kwargs)

                endTime = time.time()
                totalTime = endTime-initTime

                args[0].logger.info("Elapsed time is %f seconds." % totalTime)
                args[0].logger.info()

                return val

            return wrapper

        return metadata_decor



# from ..modules.node_features import NodeFeatures
# from ..modules import word_processor as wdprc
# from ..modules import clean_data as clean
# from ..modules import file_ops as flop
# from ..modules import logger as logger
# from ..modules.logger import Process
# from collections import OrderedDict
# from math import ceil
# import random

from text4gcn.modules.node_features import NodeFeatures
from text4gcn.modules import word_processor as wdprc
from text4gcn.modules import clean_data as clean
from text4gcn.modules import file_ops as flop
from text4gcn.modules import logger as logger
from text4gcn.modules.logger import Process
from collections import OrderedDict
from math import ceil
import random


class TextPipeline():
    """
    A class to represent a person.

    ...

    Attributes
    ----------
    name : str
        first name of the person
    surname : str
        family name of the person
    age : int
        age of the person

    Methods
    -------
    info(additional=""):
        Prints the person's name and age.
    """

    def __init__(
        self,
        dataset_name: str,
        rare_count: int,
        dataset_path: str,
        language: str
    ):
        self.logger = logger.PrintLog()
        self.flop = flop.FileOps(logger=self.logger)
        self.dataset_name = dataset_name
        self.dataset_path = dataset_path
        self.rare_count = rare_count
        self.language = language

    def save_history(self, hist):
        self.flop.create_dir(
            dir_path=f'{self.dataset_path}/log',
            overwrite=False
        )

        with open(f'{self.dataset_path}/log/{self.dataset_name}_dataset.txt', 'w') as my_file:
            my_file.writelines(hist)

    def execute(self):
        self.logger.info(clean.create_title("NLTK Configuration"))
        clean.config_nltk()
        self.logger.info()

        self.clean_data()
        self.shuffle_data()
        self.prepare_words()
        self.build_node_features(
            validation_ratio=0.10,
            use_predefined_word_vectors=False
        )

        hist = self.logger.log_history()
        self.save_history(hist)

    @Process.log("CLEANED DATA: Removed rare & stop-words.")
    def clean_data(self):

        corpus_path = f"{self.dataset_path}/{self.dataset_name}.txt"
        corpus_cleaned = f"{self.dataset_path}/{self.dataset_name}.cleaned"

        # Checkers
        # flop.check_data_set(data_set_name=self.dataset_name, all_data_set_names=self.config.data_sets)
        self.flop.check_paths(self.dataset_path)

        self.flop.create_dir(
            dir_path=corpus_cleaned,
            overwrite=False)

        docs_of_words = [clean.clean_str(line.strip().decode(
            'latin1')).split() for line in open(corpus_path, 'rb')]

        word_counts = clean.extract_word_counts(docs_of_words=docs_of_words)

        stop_words = clean.retrieve_stop_words(language=self.language)

        # TODO: IF
        docs_of_words = clean.remove_stop_words(
            docs_of_words,
            stop_words=stop_words
        )

        # TODO: IF
        docs_of_words = clean.remove_rare_words(
            docs_of_words,
            word_counts=word_counts,
            rare_count=self.rare_count
        )

        docs_of_words = clean.glue_lines(
            lines_of_words=docs_of_words,
            glue_str=' ',
            with_strip=True
        )

        self.flop.write_iterable_to_file(
            an_iterable=docs_of_words,
            file_path=f'{corpus_cleaned}/{self.dataset_name}.txt',
            file_mode='w'
        )

        self.logger.info(f"Rare-Count = <{self.rare_count}>")

    @Process.log("SHUFFLED DATA: Corpus documents shuffled.")
    def shuffle_data(self):

        ds_corpus = f"{self.dataset_path}/{self.dataset_name}.cleaned/{self.dataset_name}.txt"
        ds_corpus_meta = f"{self.dataset_path}/{self.dataset_name}.meta"

        ds_corpus_shuffled = f"{self.dataset_path}/{self.dataset_name}.shuffled/{self.dataset_name}.txt"
        ds_corpus_shuffled_train_idx = f"{self.dataset_path}/{self.dataset_name}.shuffled/{self.dataset_name}.train"
        ds_corpus_shuffled_test_idx = f"{self.dataset_path}/{self.dataset_name}.shuffled/{self.dataset_name}.test"
        ds_corpus_shuffled_meta = f"{self.dataset_path}/{self.dataset_name}.shuffled/{self.dataset_name}.meta"

        # Checkers
        #check_data_set(data_set_name=ds_name, all_data_set_names=cfg.data_sets)
        #check_paths(ds_corpus_meta, corpus_cleaned)

        # Create dirs if not exist
        self.flop.create_dir(
            dir_path=f"{self.dataset_path}/{self.dataset_name}.shuffled/",
            overwrite=False
        )

        all_doc_meta_list, train_doc_meta_list, test_doc_meta_list = self.flop.load_corpus_meta(
            corpus_meta_path=ds_corpus_meta
        )

        cleaned_doc_lines = [line.strip() for line in open(ds_corpus, 'r')]

        # Shuffle train ids and write to file
        train_doc_meta_ids = [all_doc_meta_list.index(
            train_doc_meta) for train_doc_meta in train_doc_meta_list]

        random.shuffle(train_doc_meta_ids)

        self.flop.write_iterable_to_file(
            an_iterable=train_doc_meta_ids,
            file_path=ds_corpus_shuffled_train_idx,
            file_mode='w'
        )

        # Shuffle test ids and write to file
        test_doc_meta_ids = [all_doc_meta_list.index(
            test_doc_meta) for test_doc_meta in test_doc_meta_list]

        random.shuffle(test_doc_meta_ids)

        self.flop.write_iterable_to_file(
            an_iterable=test_doc_meta_ids,
            file_path=ds_corpus_shuffled_test_idx,
            file_mode='w'
        )

        all_doc_meta_ids = train_doc_meta_ids + test_doc_meta_ids
        # Write shuffled meta to file
        shuffled_doc_meta_list = [all_doc_meta_list[all_doc_meta_id]
                                  for all_doc_meta_id in all_doc_meta_ids]

        self.flop.write_iterable_to_file(
            an_iterable=shuffled_doc_meta_list,
            file_path=ds_corpus_shuffled_meta,
            file_mode='w'
        )

        # Write shuffled document files to file
        shuffled_doc_lines = [cleaned_doc_lines[all_doc_meta_id]
                              for all_doc_meta_id in all_doc_meta_ids]

        self.flop.write_iterable_to_file(
            an_iterable=shuffled_doc_lines,
            file_path=ds_corpus_shuffled,
            file_mode='w'
        )

    @Process.log("PREPARED WORDS: Vocabulary & word-vectors extracted.")
    def prepare_words(self):

        ds_corpus = f"{self.dataset_path}/{self.dataset_name}.shuffled/{self.dataset_name}.txt"

        # Checkers
        #check_data_set(data_set_name=ds_name, all_data_set_names=cfg.data_sets)
        # check_paths(ds_corpus)
        self.flop.check_paths(self.dataset_path)

        # Create output directories
        #create_dir(dir_path=cfg.corpus_shuffled_vocab_dir, overwrite=False)
        #create_dir(dir_path=cfg.corpus_shuffled_word_vectors_dir, overwrite=False)

        ds_corpus_vocabulary = f"{self.dataset_path}/{self.dataset_name}.shuffled/{self.dataset_name}.vocab"
        ds_corpus_word_vectors = f"{self.dataset_path}/{self.dataset_name}.shuffled/{self.dataset_name}.word_vectors"

        # ###################################################

        # Build vocabulary
        docs_of_words_generator = (line.split() for line in open(ds_corpus))

        vocabulary = wdprc.extract_vocabulary(
            docs_of_words=docs_of_words_generator
        )

        self.flop.write_iterable_to_file(
            an_iterable=vocabulary,
            file_path=ds_corpus_vocabulary,
            file_mode='w'
        )

        # # Extract word definitions
        word_definitions = wdprc.extract_word_definitions(
            vocabulary=vocabulary
        )

        # # Extract & Dump word vectors
        word_vectors = wdprc.extract_tf_idf_word_vectors(
            word_definitions=word_definitions, max_features=1000
        )

        word_to_word_vectors_dict = wdprc.word_to_vectors(
            vocabulary, word_vectors
        )

        self.flop.write_picke(
            word_to_word_vectors_dict,
            ds_corpus_word_vectors
        )

    @Process.log("EXTRACTED NODE FEATURES: x, y, tx, ty, allx, ally.")
    def build_node_features(self, validation_ratio: float, use_predefined_word_vectors: bool):

        ds_corpus = f"{self.dataset_path}/{self.dataset_name}.shuffled/{self.dataset_name}.txt"
        ds_corpus_meta = f"{self.dataset_path}/{self.dataset_name}.shuffled/{self.dataset_name}.meta"
        ds_corpus_vocabulary = f"{self.dataset_path}/{self.dataset_name}.shuffled/{self.dataset_name}.vocab"
        ds_corpus_train_idx = f"{self.dataset_path}/{self.dataset_name}.shuffled/{self.dataset_name}.train"
        ds_corpus_test_idx = f"{self.dataset_path}/{self.dataset_name}.shuffled/{self.dataset_name}.test"

        # output directory of node features
        dir_corpus_node_features = f"{self.dataset_path}/{self.dataset_name}.node_features"

        # checkers
        self.flop.check_paths(ds_corpus, ds_corpus_meta, ds_corpus_vocabulary)
        self.flop.check_paths(ds_corpus_train_idx, ds_corpus_train_idx)

        # Create output directory of node features
        self.flop.create_dir(
            dir_path=dir_corpus_node_features,
            overwrite=False
        )

        # Adjust train size, for different training rates, for example: use 90% of training set
        real_train_size = len(open(ds_corpus_train_idx).readlines())
        adjusted_train_size = ceil(real_train_size * (1.0 - validation_ratio))
        test_size = len(open(ds_corpus_test_idx).readlines())

        nfeatures = NodeFeatures(logger=self.logger)

        # =============================================================== TESTE
        # # Extract word_vectors and word_embedding_dimension
        # if use_predefined_word_vectors:
        #     #ds_corpus_word_vectors = cfg.corpus_shuffled_word_vectors_dir + ds_name + '.word_vectors'
        #     ds_corpus_word_vectors = f"{self.dataset_path}/{self.dataset_name}.shuffled/{self.dataset_name}.word_vectors"
        #     # ds_corpus_word_vectors =  'glove.6B.300d.txt'  # Alternatively, you can use GLOVE word-embeddings
        #     word_vectors, word_emb_dim = nfeatures.load_word_to_word_vectors(
        #         path=ds_corpus_word_vectors)
        # else:
        #     word_vectors, word_emb_dim = OrderedDict(), 300
        # ===============================================================

        # TODO: parametrize
        word_vectors, word_emb_dim = OrderedDict(), 300

        # Extract Vocabulary
        vocabulary = open(ds_corpus_vocabulary).read().splitlines()
        # Extract Meta List
        doc_meta_list = open(file=ds_corpus_meta, mode='r').read().splitlines()
        # Extract Document Labels
        doc_labels = nfeatures.extract_doc_labels(
            ds_corpus_meta_file=ds_corpus_meta)
        # Extract Documents of Words
        docs_of_words = [line.split() for line in open(file=ds_corpus)]

        # Extract mean document word vectors and one hot labels of train-set
        # The feature vectors of the training instances as scipy.sparse.csr.csr_matrix object;
        x = nfeatures.compute_x(
            docs_of_words,
            adjusted_train_size,
            word_emb_dim,
            w_vectors=word_vectors
        )

        # Extract mean document word vectors and one hot labels of test-set
        # The one-hot labels of the labeled training instances as numpy.ndarray object;
        y = nfeatures.compute_y(
            doc_meta_list,
            train_size=adjusted_train_size,
            doc_labels=doc_labels
        )

        # The feature vectors of the test instances as scipy.sparse.csr.csr_matrix object;
        tx = nfeatures.compute_tx(
            docs_of_words,
            test_size,
            real_train_size,
            word_emb_dim,
            w_vectors=word_vectors
        )

        # The one-hot labels of the test instances as numpy.ndarray object;
        ty = nfeatures.compute_ty(
            doc_meta_list,
            test_size=test_size,
            real_train_size=real_train_size,
            doc_labels=doc_labels
        )

        # Extract doc_features + word_features
        # The feature vectors of both labeled and unlabeled training instances (a
        # superset of ind.dataset_str.x) as scipy.sparse.csr.csr_matrix object;
        allx = nfeatures.compute_allx(
            docs_of_words,
            real_train_size,
            vocabulary,
            word_vectors,
            emb_dim=word_emb_dim
        )

        # The labels for instances in ind.dataset_str.allx as numpy.ndarray object;
        ally = nfeatures.compute_ally(
            doc_meta_list,
            real_train_size,
            doc_labels,
            vocab_size=len(vocabulary)
        )

        # Dump node features matrices to files
        node_feature_matrices = {
            "x": x,
            "y": y,
            "tx": tx,
            "ty": ty,
            "allx": allx,
            "ally": ally
        }

        nfeatures.dump_node_features(
            directory=dir_corpus_node_features,
            ds=self.dataset_name,
            node_features_dict=node_feature_matrices
        )

        self.logger.info(f"x.shape={x.shape}, y.shape={y.shape}")
        self.logger.info(f"tx.shape={tx.shape}, ty.shape={ty.shape}")
        self.logger.info(f"allx.shape={allx.shape}, ally.shape={ally.shape}")



from os.path import isabs
from os import getcwd


class PreProcessingConfig():

    def __init__(self):

        self.dataset_name = None
        self.rare_count = None
        self.dataset_path = None
        self.language = None


        # # List of Valid Data-sets
        # self.data_sets = None
        # # List of Valid Adjacency-sets
        # self.adjacency_sets = None
        # # Extension of data-sets, e.g. "txt"
        # self.data_set_extension = None
        # # Original Corpus Directory
        # self.corpus_dir = None
        # # Cleaned Corpus Directory
        # self.corpus_cleaned_dir = None
        # # Original Meta Directory of Corpus
        # self.corpus_meta_dir = None
        # # Shuffled Corpus Directory
        # self.corpus_shuffled_dir = None
        # # Train and Test Index of Shuffled Corpus
        # self.corpus_shuffled_split_index_dir = None
        # # Meta Directory of Shuffled Corpus
        # self.corpus_shuffled_meta_dir = None
        # # Vocabulary of Shuffled Corpus
        # self.corpus_shuffled_vocab_dir = None
        # # Word-Vectors of Shuffled Corpus
        # self.corpus_shuffled_word_vectors_dir = None
        # # Node Features (x,y,tx,ty,allx) of Shuffled Corpus
        # self.corpus_shuffled_node_features_dir = None
        # # Adjacency Matrix (adj) of Shuffled Corpus
        # self.corpus_shuffled_adjacency_dir = None
        # self.core_nlp_path = None
        # self.liwc_path = None

    

    # def make_path_absolute(self, a_path: str) -> str:

    #     if isabs(a_path):
    #         print(f'[WARN] Path:{a_path} is already absolute.')
    #         return a_path

    #     current_working_dir = getcwd()

    #     return f'{current_working_dir}/{a_path}'

    # def build(self) -> 'PreProcessingConfig':

    #     self.corpus_dir = self.make_path_absolute(
    #         self.corpus_dir)

    #     self.corpus_cleaned_dir = self.make_path_absolute(
    #         self.corpus_cleaned_dir)

    #     self.corpus_meta_dir = self.make_path_absolute(
    #         self.corpus_meta_dir)

    #     self.corpus_shuffled_dir = self.make_path_absolute(
    #         self.corpus_shuffled_dir)

    #     self.corpus_shuffled_split_index_dir = self.make_path_absolute(
    #         self.corpus_shuffled_split_index_dir)

    #     self.corpus_shuffled_meta_dir = self.make_path_absolute(
    #         self.corpus_shuffled_meta_dir)

    #     self.corpus_shuffled_vocab_dir = self.make_path_absolute(
    #         self.corpus_shuffled_vocab_dir)

    #     self.corpus_shuffled_word_vectors_dir = self.make_path_absolute(
    #         self.corpus_shuffled_word_vectors_dir)

    #     self.corpus_shuffled_node_features_dir = self.make_path_absolute(
    #         self.corpus_shuffled_node_features_dir)

    #     self.corpus_shuffled_adjacency_dir = self.make_path_absolute(
    #         self.corpus_shuffled_adjacency_dir)

    #     self.core_nlp_path = self.make_path_absolute(
    #         self.core_nlp_path)

    #     return self



from datetime import timedelta, datetime
from tabulate import tabulate
import scipy.sparse as sp
import pickle as pkl
import torch as th
import numpy as np
import random
import time


class LoadData:

    def __init__(self, path, dataset, builder):
        self.path = path
        self.dataset = dataset
        self.builder = builder

    def parse_index_file(self, filename):
        """Parse index file."""
        return [int(line.strip()) for line in open(filename)]

    def sample_mask(self, idx, row_length: int):
        """Create mask."""
        mask = np.zeros(row_length)
        mask[idx] = 1
        return np.array(mask, dtype=np.bool)

    def load_corpus(self):
        """
        Loads input corpus from gcn/data directory

        ind.dataset_str.x => the feature vectors of the training docs as scipy.sparse.csr.csr_matrix object;
        ind.dataset_str.tx => the feature vectors of the test docs as scipy.sparse.csr.csr_matrix object;
        ind.dataset_str.allx => the feature vectors of both labeled and unlabeled training docs/words
            (a superset of ind.dataset_str.x) as scipy.sparse.csr.csr_matrix object;
        ind.dataset_str.y => the one-hot labels of the labeled training docs as numpy.ndarray object;
        ind.dataset_str.ty => the one-hot labels of the test docs as numpy.ndarray object;
        ind.dataset_str.ally => the labels for instances in ind.dataset_str.allx as numpy.ndarray object;
        ind.dataset_str.adj => adjacency matrix of word/doc nodes as scipy.sparse.csr.csr_matrix object;
        ind.dataset_str.train.index => the indices of training docs in original doc list.

        All objects above must be saved using python pickle module.

        :param dataset_str: Dataset name
        :return: All data input files loaded (as well the training/test data).
        """

        node_feature_names = ['x', 'y', 'tx', 'ty', 'allx', 'ally']
        node_features = []
        for node_feature_name in node_feature_names:
            # with open(f"{self.path}/{self.dataset}.node_features/ind.{self.builder}.{self.dataset}.{node_feature_name}", 'rb') as f:
            with open(f"{self.path}/{self.dataset}.node_features/ind.{self.dataset}.{node_feature_name}", 'rb') as f:
                node_features.append(pkl.load(f, encoding='latin1'))

        with open(f"{self.path}/{self.dataset}.adjacency/ind.{self.builder}.{self.dataset}.adj", 'rb') as f:
            adj = pkl.load(f, encoding='latin1')

        x, y, tx, ty, allx, ally = tuple(node_features)

        features = sp.vstack((allx, tx)).tolil()
        labels = np.vstack((ally, ty))

        print("\nShow info about load data:\n")

        print(tabulate([
            ["x", str(x.shape)],
            ["y", str(y.shape)],
            ["tx", str(tx.shape)],
            ["ty", str(ty.shape)],
            ["allx", str(allx.shape)],
            ["ally", str(ally.shape)],
            ["adj", str(adj.shape)],
            ["Features", str(features.shape)],
            ["Labels", str(labels.shape)],
        ], ["Data", "Shape"]))

        train_idx_orig = self.parse_index_file(
            f"{self.path}/{self.dataset}.shuffled/{self.dataset}.train")

        train_size = len(train_idx_orig)
        val_size = train_size - x.shape[0]
        test_size = tx.shape[0]

        # This part of the index corresponds to the value after vstack
        idx_train = range(len(y))
        idx_val = range(len(y), len(y) + val_size)
        idx_test = range(allx.shape[0], allx.shape[0] + test_size)

        # sample_mask plays the role of dividing the data area, the following three lines divide the doc part
        train_mask = self.sample_mask(idx_train, labels.shape[0])
        val_mask = self.sample_mask(idx_val, labels.shape[0])
        test_mask = self.sample_mask(idx_test, labels.shape[0])

        y_train = np.zeros(labels.shape)
        y_val = np.zeros(labels.shape)
        y_test = np.zeros(labels.shape)
        # Numpy bool slice operation
        y_train[train_mask, :] = labels[train_mask, :]
        y_val[val_mask, :] = labels[val_mask, :]
        y_test[test_mask, :] = labels[test_mask, :]

        # Convert asymmetric adjacency matrix to symmetric adjacency matrix
        adj = adj + adj.T.multiply(adj.T > adj) - adj.multiply(adj.T > adj)

        return adj, features, y_train, y_val, y_test, train_mask, val_mask, test_mask, train_size, test_size


def unique_name():
    return datetime.utcnow().strftime('%Y%m%d-%H%M%S%f')[:-3]


def get_train_test(target_fn):
    train_lst = []
    test_lst = []
    with read_file(target_fn, mode="r") as fin:
        for indx, item in enumerate(fin):
            if item.split("\t")[1] in {"train", "training", "20news-bydate-train"}:
                train_lst.append(indx)
            else:
                test_lst.append(indx)

    return train_lst, test_lst


def read_file(path, mode='r', encoding=None):
    if mode not in {"r", "rb"}:
        raise ValueError("only read")
    return open(path, mode=mode, encoding=encoding)


def return_seed(nums=10):
    return random.sample(range(100000), nums)


def get_time_dif(start_time):
    """Obter tempo decorrido"""
    end_time = time.time()
    time_dif = end_time - start_time
    return timedelta(seconds=int(round(time_dif)))


def macro_f1(pred, targ, num_classes=None):
    pred = th.max(pred, 1)[1]
    tp_out = []
    fp_out = []
    fn_out = []
    if num_classes is None:
        num_classes = sorted(set(targ.cpu().numpy().tolist()))
    else:
        num_classes = range(num_classes)
    for i in num_classes:
        # A previsão é i, e o rótulo é de fato i
        tp = ((pred == i) & (targ == i)).sum().item()
        # A previsão é i, mas o rótulo não é i
        fp = ((pred == i) & (targ != i)).sum().item()
        # A previsão não é i, mas o rótulo é i
        fn = ((pred != i) & (targ == i)).sum().item()
        tp_out.append(tp)
        fp_out.append(fp)
        fn_out.append(fn)

    eval_tp = np.array(tp_out)
    eval_fp = np.array(fp_out)
    eval_fn = np.array(fn_out)

    precision = eval_tp / (eval_tp + eval_fp)
    precision[np.isnan(precision)] = 0
    precision = np.mean(precision)

    recall = eval_tp / (eval_tp + eval_fn)
    recall[np.isnan(recall)] = 0
    recall = np.mean(recall)

    f1 = 2 * (precision * recall) / (precision + recall)
    return f1, precision, recall


def accuracy(pred, targ):
    pred = th.max(pred, 1)[1]
    return ((pred == targ).float()).sum().item() / targ.size()[0]


def sparse_to_tensor(sparse_matrix):
    values = sparse_matrix.data
    indices = np.vstack((sparse_matrix.row, sparse_matrix.col))
    i = th.LongTensor(indices)
    v = th.FloatTensor(values)
    shape = sparse_matrix.shape
    return th.sparse.FloatTensor(i, v, th.Size(shape))


def preprocess_features(features):
    """Row-normalize feature matrix and convert to tuple representation"""
    rowsum = np.array(features.sum(1))
    r_inv = np.power(rowsum, -1).flatten()
    r_inv[np.isinf(r_inv)] = 0.
    r_mat_inv = sp.diags(r_inv)
    features = r_mat_inv.dot(features)
    # return sparse_to_tuple(features)
    return features.A


def normalize_adj(adj):
    """Symmetrically normalize adjacency matrix."""
    adj = sp.coo_matrix(adj)
    rowsum = np.array(adj.sum(1))
    d_inv_sqrt = np.power(rowsum, -0.5).flatten()
    d_inv_sqrt[np.isinf(d_inv_sqrt)] = 0.
    d_mat_inv_sqrt = sp.diags(d_inv_sqrt)
    return adj.dot(d_mat_inv_sqrt).transpose().dot(d_mat_inv_sqrt).tocoo()


def preprocess_adj(adj):
    """Preprocessing of adjacency matrix for simple GCN model and conversion to tuple representation."""
    adj_normalized = normalize_adj(adj + sp.eye(adj.shape[0]))
    # return sparse_to_tuple(adj_normalized)
    return adj_normalized.A



import requests
import os


def list():
    return ['R8', 'R52', 'AG_NEWS']


class Datasets():
    def __init__(self, file_url: str, path: str, file_name: str) -> None:
        self.file_url = file_url
        self.path = path
        self.file_name = file_name

    def download(self):
        # URL of the image to be downloaded is defined as image_url
        r = requests.get(self.file_url)  # create HTTP response object

        #file_path = f'{os.getcwd()}/{path}'
        absp = os.path.abspath(self.path)
        file_path = f'{absp}/{self.file_name}'

        # send a HTTP request to the server and save
        # the HTTP response in a response object called r
        with open(file_path, 'wb') as f:

            # write the contents of the response (r.content)
            # to a new file in binary mode.
            f.write(r.content)

    def download_big(self):
        r = requests.get(self.file_url, stream=True)

        absp = os.path.abspath(self.path)
        file_path = f'{absp}/{self.file_name}'

        with open(file_path, "wb") as pdf:
            for chunk in r.iter_content(chunk_size=1024):

                # writing one chunk at a time to file
                if chunk:
                    pdf.write(chunk)


def R8(path: str):
    """R8 Dataset
    For additional details refer to http://ai.stanford.edu/~amaas/data/sentiment/
    Number of lines per split:
        - train: 25000
        - test: 25000
    Args:
        root: Directory where the datasets are saved. Default: os.path.expanduser('~/.torchtext/cache')
        split: split or splits to be returned. Can be a string or tuple of strings. Default: (`train`, `test`)
    :returns: DataPipe that yields tuple of label (1 to 2) and text containing the movie review
    :rtype: (int, str)
    """
    ds = Datasets(
        file_url="https://meriatblog.blob.core.windows.net/public/text4gcn/R8.txt",
        file_name="R8.txt",
        path=path)
    ds.download()

    ds = Datasets(
        file_url="https://meriatblog.blob.core.windows.net/public/text4gcn/R8.meta",
        file_name="R8.meta",
        path=path)
    ds.download()


def R52(path: str):
    """R52 Dataset
    For additional details refer to http://ai.stanford.edu/~amaas/data/sentiment/
    Number of lines per split:
        - train: 25000
        - test: 25000
    Args:
        root: Directory where the datasets are saved. Default: os.path.expanduser('~/.torchtext/cache')
        split: split or splits to be returned. Can be a string or tuple of strings. Default: (`train`, `test`)
    :returns: DataPipe that yields tuple of label (1 to 2) and text containing the movie review
    :rtype: (int, str)
    """
    ds = Datasets(
        file_url="https://meriatblog.blob.core.windows.net/public/text4gcn/R52.txt",
        file_name="R52.txt",
        path=path)
    ds.download()

    ds = Datasets(
        file_url="https://meriatblog.blob.core.windows.net/public/text4gcn/R52.meta",
        file_name="R52.meta",
        path=path)
    ds.download()


def AG_NEWS(path: str):
    """AG_NEWS Dataset
    For additional details refer to http://ai.stanford.edu/~amaas/data/sentiment/
    Number of lines per split:
        - train: 25000
        - test: 25000
    Args:
        root: Directory where the datasets are saved. Default: os.path.expanduser('~/.torchtext/cache')
        split: split or splits to be returned. Can be a string or tuple of strings. Default: (`train`, `test`)
    :returns: DataPipe that yields tuple of label (1 to 2) and text containing the movie review
    :rtype: (int, str)
    """
    ds = Datasets(
        file_url="https://meriatblog.blob.core.windows.net/public/text4gcn/20AG_NEWS.txt",
        file_name="20AG_NEWS.txt",
        path=path)
    ds.download()

    ds = Datasets(
        file_url="https://meriatblog.blob.core.windows.net/public/text4gcn/20AG_NEWS.meta",
        file_name="20AG_NEWS.meta",
        path=path)
    ds.download()



from text4gcn.modules import file_ops as flop
from text4gcn.modules import logger as logger
from text4gcn.modules import adjacency as adj
from text4gcn.modules.logger import Process
from scipy.sparse import csr_matrix
import pickle


class FrequencyAdjacency():

    def __init__(self, dataset_name, dataset_path):
        self.dataset_name = dataset_name
        self.dataset_path = dataset_path
        self.logger = logger.PrintLog()
        self.flop = flop.FileOps(logger=self.logger)

    def save_history(self, hist):
        self.flop.create_dir(
            dir_path=f'{self.dataset_path}/log', overwrite=False)
        with open(f'{self.dataset_path}/log/{self.dataset_name}_dataset.txt', 'a') as my_file:
            my_file.writelines(hist)

    def build(self):
        self.builds()
        hist = self.logger.log_history()
        self.save_history(hist)

    @Process.log("EXTRACTED ADJACENCY MATRIX: Heterogenous doc-word adjacency matrix.")
    def builds(self):

        corpus_path = f"{self.dataset_path}/{self.dataset_name}"
        ds_corpus = f'{corpus_path}.shuffled/{self.dataset_name}.txt'
        ds_corpus_vocabulary = f'{corpus_path}.shuffled/{self.dataset_name}.vocab'
        ds_corpus_train_idx = f'{corpus_path}.shuffled/{self.dataset_name}.train'
        ds_corpus_test_idx = f'{corpus_path}.shuffled/{self.dataset_name}.test'

        self.flop.create_dir(
            dir_path=f'{corpus_path}.adjacency', overwrite=False)

        docs_of_words = [line.split() for line in open(file=ds_corpus)]
        # Extract Vocabulary.
        vocab = open(ds_corpus_vocabulary).read().splitlines()
        # Word to its id.
        word_to_id = {word: i for i, word in enumerate(vocab)}
        # Real train-size, not adjusted.
        train_size = len(open(ds_corpus_train_idx).readlines())
        # Real test-size.
        test_size = len(open(ds_corpus_test_idx).readlines())

        windows_of_words = adj.extract_windows(
            docs_of_words=docs_of_words, window_size=20)

        self.logger.info("Calculating PMI")
        # Extract word-word weights
        rows, cols, weights = adj.extract_pmi_word_weights(
            windows_of_words, word_to_id, vocab, train_size)

        self.logger.info("Calculating TF-IDF")
        # Extract word-doc weights
        rows, cols, weights = adj.extract_tf_idf_doc_word_weights(
            rows, cols, weights, vocab, train_size, docs_of_words, word_to_id)

        adjacency_len = train_size + len(vocab) + test_size

        self.logger.info(
            f"[INFO] ({len(weights)}, ({len(rows)}, {len(cols)})), shape=({adjacency_len}, {adjacency_len})")

        adjacency_matrix = csr_matrix(
            (weights, (rows, cols)), shape=(adjacency_len, adjacency_len))

        # Dump Adjacency Matrix
        with open(f"{corpus_path}.adjacency/ind.frequency.{self.dataset_name}.adj", 'wb') as f:
            pickle.dump(adjacency_matrix, f)



# class Builder:
#     Liwc = 'liwc'
#     Frequency = 'frequency'
#     Embedding = 'embedding'
#     CosineSimilarity = 'cosine'    
#     DependencyParsing = 'dependency'
#     ConstituencyParsing = 'constituency'
    
# import enum
# from enum import StrEnum

# @enum.unique
# class Builder(StrEnum):
#     Liwc = 'liwc'
#     Frequency = 'frequency'
#     Embedding = 'embedding'
#     CosineSimilarity = 'cosine'    
#     DependencyParsing = 'dependency'
#     ConstituencyParsing = 'constituency'

from enum import Enum

class Builder(str, Enum):
    Liwc = 'liwc'
    Frequency = 'frequency'
    Embedding = 'embedding'
    CosineSimilarity = 'cosine'    
    DependencyParsing = 'dependency'
    ConstituencyParsing = 'constituency'



import torch
import torch.nn as nn


class GraphConvolution(nn.Module):
    def __init__(self, input_dim, output_dim, support, act_func=None, featureless=False, dropout_rate=0., bias=False):

        super(GraphConvolution, self).__init__()
        self.support = support
        self.featureless = featureless

        for i in range(len(self.support)):
            setattr(self, f'W{i}', nn.Parameter(
                torch.randn(input_dim, output_dim)))

        if bias:
            self.b = nn.Parameter(torch.zeros(1, output_dim))

        self.act_func = act_func
        self.dropout = nn.Dropout(dropout_rate)

    def forward(self, x):
        x = self.dropout(x)

        for i in range(len(self.support)):
            if self.featureless:
                pre_sup = getattr(self, f'W{i}')
            else:
                pre_sup = x.mm(getattr(self, f'W{i}'))

            if i == 0:
                out = self.support[i].mm(pre_sup)
            else:
                out += self.support[i].mm(pre_sup)

        if self.act_func is not None:
            out = self.act_func(out)

        self.embedding = out
        return out


class GCN(nn.Module):
    def __init__(self, input_dim, output_dim, support, dropout_rate=0., num_classes=10):
        super(GCN, self).__init__()

        # GraphConvolution
        self.layer1 = GraphConvolution(
            input_dim, output_dim, support, act_func=nn.ReLU(), featureless=True, dropout_rate=dropout_rate)

        self.layer2 = GraphConvolution(
            output_dim, num_classes, support, dropout_rate=dropout_rate)

    def forward(self, x):
        out = self.layer1(x)
        out = self.layer2(out)
        return out



# """Read the latest Real Python tutorials.
# Usage:
# ------
#     $ realpython [options] [id] [id ...]
# List the latest tutorials:
#     $ realpython
# Read one tutorial:
#     $ realpython <id>
#     where <id> is the number shown when listing tutorials.
# Read the latest tutorial:
#     $ realpython 0
# Available options are:
#     -h, --help         Show this help
#     -l, --show-links   Show links in text
# Contact:
# --------
# - https://realpython.com/contact/
# More information is available at:
# - https://pypi.org/project/realpython-reader/
# - https://github.com/realpython/reader
# Version:
# --------
# - realpython-reader v1.0.0
# """
# # Standard library imports
# import sys

# # Reader imports
# import text4gcn
# from text4gcn import builder, preprocess


# def main() -> None:
#     """Read the Real Python article feed."""
#     args = [a for a in sys.argv[1:] if not a.startswith("-")]
#     opts = [o for o in sys.argv[1:] if o.startswith("-")]

#     # Show help message
#     if "-h" in opts or "--help" in opts:
#         #preprocess.show(__doc__)
#         raise SystemExit()

#     # # Should links be shown in the text
#     # show_links = "-l" in opts or "--show-links" in opts

#     # # Get URL from config file
#     # url = reader.URL

#     # # An article ID is given, show article
#     # if args:
#     #     for article_id in args:
#     #         article = feed.get_article(article_id, links=show_links, url=url)
#     #         viewer.show(article)

#     # # No ID is given, show list of articles
#     # else:
#     #     site = feed.get_site(url=url)
#     #     titles = feed.get_titles(url=url)
#     #     viewer.show_list(site, titles)


# if __name__ == "__main__":
#     main()


from text4gcn.modules.textvec.liwc import Liwc
from text4gcn.modules import file_ops as flop
from text4gcn.modules import logger as logger
from text4gcn.modules.word_processor import *
from nltk.corpus import stopwords
import scipy.sparse as sp
from math import log
import pickle as pkl


class LiwcAdjacency():

    def __init__(self, dataset_name, dataset_path, liwc_path):
        self.dataset_name = dataset_name
        self.dataset_path = dataset_path
        self.liwc_path = liwc_path
        self.logger = logger.PrintLog()
        self.flop = flop.FileOps(logger=self.logger)

    def build_relation_pair(self, docs_of_words):
        stop_words = set(stopwords.words('english'))
        liwc = Liwc(self.liwc_path)

        errors = 0
        rela_pair_count_str = {}
        for docs_of_word in docs_of_words:
            res = []
            try:
                for word in docs_of_word:
                    if len(liwc.parse(word)):
                        res.append(
                            {'token': word, 'class': liwc.get_class(word)[1]})
                    else:
                        res.append({'token': word, 'class': None})

            except Exception as e:
                errors += 1

            res_len = len(res)
            for i in range(res_len):
                if i+1 == res_len:
                    continue

                if res[i]['class'] is None or res[i + 1]['class'] is None:
                    continue

                if res[i]['class'] == res[i+1]['class']:
                    continue

                if res[i]['token'] in stop_words or res[i+1]['token'] in stop_words:
                    continue

                word_pair_str = res[i]['token'] + ',' + res[i+1]['token']
                if word_pair_str in rela_pair_count_str:
                    rela_pair_count_str[word_pair_str] += 1
                else:
                    rela_pair_count_str[word_pair_str] = 1
                # two orders
                word_pair_str = res[i+1]['token'] + ',' + res[i]['token']
                if word_pair_str in rela_pair_count_str:
                    rela_pair_count_str[word_pair_str] += 1
                else:
                    rela_pair_count_str[word_pair_str] = 1
        print(f'[INFO] Error in Dependency Parse: {errors}')
        return rela_pair_count_str

    def compute_weights_with_PMI(self, docs_of_words, rela_pair_count_str, word_window_freq, train_size, word_id_map, min_count1, max_count1, count_mean1, count_std1, row, col):
        weight = []
        errors = 0
        num_window = len(docs_of_words)
        for key, count in rela_pair_count_str.items():
            try:
                temp = key.split(',')
                i = temp[0]
                j = temp[1]

                if i in word_window_freq and j in word_window_freq:
                    word_freq_i = word_window_freq[i]
                    word_freq_j = word_window_freq[j]
                    pmi = log(
                        1.0
                        * count
                        / num_window
                        / (1.0 * word_freq_i * word_freq_j / num_window ** 2)
                    )
                    if pmi <= 0:
                        continue

                    row.append(train_size + word_id_map[i])
                    col.append(train_size + word_id_map[j])
                    if key in rela_pair_count_str:
                        wei = (rela_pair_count_str[key] -
                               min_count1) / (max_count1 - min_count1)
                        wei = (
                            rela_pair_count_str[key]-count_mean1) / count_std1
                        weight.append(wei)
            except:
                errors += 1

        print(f'[INFO] Error in Compute Weights: {errors}')
        return weight

    def compute_weights(self, docs_of_words, rela_pair_count_str, word_window_freq, train_size, word_id_map, vocab_size, word_doc_freq, vocab):
        row = []
        col = []

        min_count1, max_count1, count_mean1, count_std1 = relation_pair_statitcs(
            rela_pair_count_str)

        # compute weights PMI
        weight = self.compute_weights_with_PMI(docs_of_words, rela_pair_count_str, word_window_freq,
                                               train_size, word_id_map, min_count1, max_count1, count_mean1, count_std1, row, col)

        # doc word frequency
        doc_word_freq = get_doc_word_freq(docs_of_words, word_id_map)

        weight_tfidf = get_weight_tfidf(
            docs_of_words, word_id_map, doc_word_freq, train_size, vocab_size, word_doc_freq, vocab, row, col)

        return row, col, weight, weight_tfidf

    def build(self):

        # corpus_path = f"{self.dataset_path}/{self.dataset_name}"
        # ds_corpus = f'{corpus_path}_clean/{self.dataset_name}.txt'
        # ds_corpus_vocabulary = f'{corpus_path}_clean/{self.dataset_name}.vocab'
        # ds_corpus_train_idx = f'{corpus_path}_clean/{self.dataset_name}.train'
        # ds_corpus_test_idx = f'{corpus_path}_clean/{self.dataset_name}.test'

        corpus_path = f"{self.dataset_path}/{self.dataset_name}"

        ds_corpus = f'{corpus_path}.shuffled/{self.dataset_name}.txt'
        ds_corpus_vocabulary = f'{corpus_path}.shuffled/{self.dataset_name}.vocab'
        ds_corpus_train_idx = f'{corpus_path}.shuffled/{self.dataset_name}.train'
        ds_corpus_test_idx = f'{corpus_path}.shuffled/{self.dataset_name}.test'

        # checkers
        #check_data_set(data_set_name=ds_name, all_data_set_names=cfg.data_sets)
        #check_paths(ds_corpus, ds_corpus_vocabulary, ds_corpus_train_idx, ds_corpus_test_idx)

        #create_dir(dir_path=cfg.corpus_shuffled_adjacency_dir + "/linguistic_inquiry", overwrite=False)
        self.flop.create_dir(
            dir_path=f'{corpus_path}.adjacency', overwrite=False)

        docs_of_words = [line.split() for line in open(file=ds_corpus)]
        # Extract Vocabulary.
        vocab = open(ds_corpus_vocabulary).read().splitlines()
        # Real train-size, not adjusted.
        train_size = len(open(ds_corpus_train_idx).readlines())
        # Real test-size.
        test_size = len(open(ds_corpus_test_idx).readlines())

        word_doc_freq = define_word_doc_freq(docs_of_words)

        vocab_size = len(vocab)
        word_id_map = {}
        id_word_map = {}
        for i in range(vocab_size):
            word_id_map[vocab[i]] = i
            id_word_map[i] = vocab[i]

        word_window_freq = build_word_window_freq(docs_of_words)

        rela_pair_count_str = self.build_relation_pair(docs_of_words)

        row, col, weight, weight_tfidf = self.compute_weights(docs_of_words, rela_pair_count_str,
                                                              word_window_freq, train_size, word_id_map,
                                                              vocab_size, word_doc_freq, vocab)

        # =============================================================
        weight += weight_tfidf
        node_size = train_size + vocab_size + test_size

        self.logger.info(
            f"({len(weight)}, ({len(row)}, {len(col)})), shape=({node_size}, {node_size})")

        adj = sp.csr_matrix((weight, (row, col)), shape=(node_size, node_size))
        # =============================================================
        # Dump Adjacency Matrix
        with open(f"{corpus_path}.adjacency/ind.liwc.{self.dataset_name}.adj", 'wb') as f:
            pkl.dump(adj, f)



# from time import time

# class TextGraph():

#     def __init__(self):
#         pass

#     def execute(self):
#         t1 = time()

#         corpus_cleaned = f"{self.dataset_path}/{self.dataset_name}_clean.txt"

#         docs_of_words = [line.split() for line in open(file=corpus_cleaned)]
       
    


from .frequency_adj import FrequencyAdjacency
from .embedding_adj import EmbeddingAdjacency
from .linguistic_inquiry_adj import LiwcAdjacency
from .cosine_similarity_adj import CosineSimilarityAdjacency
from .dependency_parsing_adj import DependencyParsingAdjacency
from .constituency_parsing_adj import ConstituencyParsingAdjacency



from gensim.models import Word2Vec
from typing import List
import numpy as np
import os


def train_word2vec(
    save_dir: str,
    document_list: List[str],
    num_epochs: int,
    embedding_dimension: int,
    training_regime: int,
) -> Word2Vec:
    """Training regime:
    - 1 for skip-gram;
    - otherwise (0) CBOW.
    """
    # # Tokenize
    # cv = CountVectorizer(tokenizer=lambda text: tokenize_prune_stem(
    #     text, stemming_map=stemming_map))
    # cv_tokenizer = cv.build_tokenizer()
    # document_list = [cv_tokenizer(document) for document in document_list]

    # Convert to TaggedDocument and train
    print('[INFO] Training Word2Vec...')
    model = Word2Vec(document_list, vector_size=embedding_dimension,
                     window=5, workers=4, sg=training_regime, min_count=1)

    if save_dir is not None:
        model.save(os.path.join(save_dir, 'word2vec.model'))

    return model


def infer_word2vec_embeddings(model: Word2Vec, word_list: List[str]) -> np.ndarray:
    """
    NOTE: Inference is not deterministic therefore representations will vary between calls
    Returns a 2D array with shape (num_words, embedding_dimension)
    """
    print('Infering word embeddings..')
    return np.array([model[word] for word in word_list])



from typing import List, Dict, Tuple, Iterable, Any
from scipy.spatial.distance import cosine
from collections import Counter
from tqdm import tqdm
from math import log
import numpy as np
import itertools
import pickle


def flatten_nested_iterables(iterables_of_iterables: Iterable[Iterable[Any]]) -> Iterable[Any]:
    return [item for sublist in iterables_of_iterables for item in sublist]


def extract_word_to_doc_ids(docs_of_words: List[List[str]]) -> Dict[str, List[int]]:
    """Extracted the document ids where unique words appeared."""
    word_to_doc_ids = {}
    for doc_id, words in enumerate(docs_of_words):
        appeared_words = set()
        for word in words:
            if word not in appeared_words:
                if word in word_to_doc_ids:
                    word_to_doc_ids[word].append(doc_id)
                else:
                    word_to_doc_ids[word] = [doc_id]
                appeared_words.add(word)
    return word_to_doc_ids


def extract_word_to_doc_counts(word_to_doc_ids: Dict[str, List[int]]) -> Dict[str, int]:
    return {word: len(doc_ids) for word, doc_ids in word_to_doc_ids.items()}


def extract_windows(docs_of_words: List[List[str]], window_size: int) -> List[List[str]]:
    """Word co-occurrence with context windows"""
    windows = []
    for doc_words in docs_of_words:
        doc_len = len(doc_words)
        if doc_len <= window_size:
            windows.append(doc_words)
        else:
            windows.extend(
                doc_words[j: j + window_size]
                for j in range(doc_len - window_size + 1)
            )

    return windows


def extract_word_counts_in_windows(windows_of_words: List[List[str]]) -> Dict[str, int]:
    """Find the total count of unique words in each window, each window is bag-of-words"""
    bags_of_words = map(set, windows_of_words)
    return Counter(flatten_nested_iterables(bags_of_words))


def extract_word_ids_pair_to_counts(windows_of_words: List[List[str]], word_to_id: Dict[str, int]) -> Dict[str, int]:
    word_ids_pair_to_counts = Counter()
    for window in windows_of_words:
        for i in range(1, len(window)):
            try:
                word_id_i = word_to_id[window[i]]
                for j in range(i):
                    word_id_j = word_to_id[window[j]]
                    if word_id_i != word_id_j:
                        word_ids_pair_to_counts.update(
                            [
                                f'{word_id_i},{word_id_j}',
                                f'{word_id_j},{word_id_i}',
                            ]
                        )
            except:
                pass

    return dict(word_ids_pair_to_counts)


def extract_pmi_word_weights(windows_of_words: List[List[str]], word_to_id: Dict[str, int], vocab: List[str],
                             train_size: int) -> Tuple[List[int], List[int], List[float]]:
    """Calculate PMI as weights"""
    weight_rows = []  # type: List[int]
    weight_cols = []  # type: List[int]
    pmi_weights = []  # type: List[float]

    num_windows = len(windows_of_words)
    word_counts_in_windows = extract_word_counts_in_windows(
        windows_of_words=windows_of_words)
    word_ids_pair_to_counts = extract_word_ids_pair_to_counts(
        windows_of_words, word_to_id)

    for word_id_pair, count in word_ids_pair_to_counts.items():
        word_ids_in_str = word_id_pair.split(',')
        word_id_i, word_id_j = int(word_ids_in_str[0]), int(word_ids_in_str[1])
        word_i, word_j = vocab[word_id_i], vocab[word_id_j]
        word_freq_i, word_freq_j = word_counts_in_windows[word_i], word_counts_in_windows[word_j]

        #pmi_score = log((1.0 * count / num_windows) / (1.0 * word_freq_i * word_freq_j / (num_windows * num_windows)))
        pmi_score = log(
            1.0
            * count
            / num_windows
            / (1.0 * word_freq_i * word_freq_j / num_windows ** 2)
        )

        if pmi_score > 0.0:
            weight_rows.append(train_size + word_id_i)
            weight_cols.append(train_size + word_id_j)
            pmi_weights.append(pmi_score)
    return weight_rows, weight_cols, pmi_weights


def extract_cosine_similarity_word_weights(vocab: List[str], train_size: int,
                                           word_vec_path: str) -> Tuple[List[int], List[int], List[float]]:
    """Calculate Cosine Similarity of Word Vectors as weights"""
    word_vectors = pickle.load(
        file=open(word_vec_path, 'rb'))  # type: Dict[str,List[float]]

    weight_rows = []  # type: List[int]
    weight_cols = []  # type: List[int]
    cos_sim_weights = []  # type: List[float]

    #len_vocab = len(vocab)
    #pbar = tqdm(len_vocab)

    for (i, word_i), (j, word_j) in tqdm(itertools.product(enumerate(vocab), enumerate(vocab))):
        # pbar.update(1)
        if word_i in word_vectors and word_j in word_vectors:
            vector_i = np.array(word_vectors[word_i])
            vector_j = np.array(word_vectors[word_j])
            similarity = 1.0 - cosine(vector_i, vector_j)
            if similarity > 0.9:
                #print(word_i, word_j, similarity)
                weight_rows.append(train_size + i)
                weight_cols.append(train_size + j)
                cos_sim_weights.append(similarity)
    # pbar.close()
    return weight_rows, weight_cols, cos_sim_weights


def extract_doc_word_ids_pair_to_counts(docs_of_words: List[List[str]], word_to_id: Dict[str, int]) -> Dict[str, int]:
    doc_word_freq = Counter()
    for doc_id, doc_words in enumerate(docs_of_words):
        for word in doc_words:
            try:
                word_id = word_to_id[word]
                doc_word_freq.update([f'{str(doc_id)},{str(word_id)}'])
            except:
                pass
    return dict(doc_word_freq)


def extract_tf_idf_doc_word_weights(
        adj_rows: List[int], adj_cols: List[int], adj_weights: List[float], vocab: List[str], train_size: int,
        docs_of_words: List[List[str]], word_to_id: Dict[str, int]) -> Tuple[List[int], List[int], List[float]]:
    """Extract Doc-Word weights with TF-IDF"""
    doc_word_ids_pair_to_counts = extract_doc_word_ids_pair_to_counts(
        docs_of_words, word_to_id)
    word_to_doc_ids = extract_word_to_doc_ids(docs_of_words=docs_of_words)
    word_to_doc_counts = extract_word_to_doc_counts(
        word_to_doc_ids=word_to_doc_ids)

    vocab_len = len(vocab)
    num_docs = len(docs_of_words)
    # percorro a lista de documentos
    for doc_id, doc_words in enumerate(docs_of_words):
        doc_word_set = set()
        # percorro as palavras do documento
        for word in doc_words:
            # se a palavra ainda não foi calculada...
            if word not in doc_word_set:
                word_id = word_to_id[word]
                word_ids_pair_count = doc_word_ids_pair_to_counts[str(
                    doc_id) + ',' + str(word_id)]

                adj_rows.append(doc_id if doc_id <
                                train_size else doc_id + vocab_len)
                adj_cols.append(train_size + word_id)

                doc_word_idf = log(
                    1.0 * num_docs / word_to_doc_counts[vocab[word_id]])
                adj_weights.append(word_ids_pair_count * doc_word_idf)
                doc_word_set.add(word)
    return adj_rows, adj_cols, adj_weights


def build_word_doc_list(docs_of_words):
    word_doc_list = {}
    for i in range(len(docs_of_words)):
        words = docs_of_words[i]
        appeared = set()
        for word in words:
            if word in appeared:
                continue
            if word in word_doc_list:
                doc_list = word_doc_list[word]
                doc_list.append(i)
                word_doc_list[word] = doc_list
            else:
                word_doc_list[word] = [i]
            appeared.add(word)
    return word_doc_list


def define_word_doc_freq(docs_of_words):
    word_doc_list = build_word_doc_list(docs_of_words)
    return {
        word: len(doc_list) for word, doc_list in word_doc_list.items()
    }


def build_word_window_freq(docs_of_words):
    word_window_freq = {}
    for window in docs_of_words:
        appeared = set()
        for i in range(len(window)):
            if window[i] in appeared:
                continue
            if window[i] in word_window_freq:
                word_window_freq[window[i]] += 1
            else:
                word_window_freq[window[i]] = 1
            appeared.add(window[i])
    return word_window_freq


def relation_pair_statitcs(rela_pair_count_str):
    max_count1 = 0.0
    min_count1 = 0.0
    count1 = []
    for key, value in rela_pair_count_str.items():
        if rela_pair_count_str[key] > max_count1:
            max_count1 = rela_pair_count_str[key]
        if value < min_count1:
            min_count1 = rela_pair_count_str[key]
        count1.append(rela_pair_count_str[key])

    count_mean1 = np.mean(count1)
    count_std1 = np.std(count1, ddof=1)
    return min_count1, max_count1, count_mean1, count_std1


def get_doc_word_freq(docs_of_words, word_id_map):
    doc_word_freq = {}
    for doc_id in range(len(docs_of_words)):
        words = docs_of_words[doc_id]
        for word in words:
            word_id = word_id_map[word]
            doc_word_str = f'{str(doc_id)},{str(word_id)}'
            if doc_word_str in doc_word_freq:
                doc_word_freq[doc_word_str] += 1
            else:
                doc_word_freq[doc_word_str] = 1
    return doc_word_freq


def get_weight_tfidf(docs_of_words, word_id_map, doc_word_freq, train_size, vocab_size, word_doc_freq, vocab, row, col):
    weight_tfidf = []
    for i in range(len(docs_of_words)):
        words = docs_of_words[i]
        doc_word_set = set()
        for word in words:
            if word in doc_word_set:
                continue
            j = word_id_map[word]
            key = f'{str(i)},{str(j)}'
            freq = doc_word_freq[key]
            if i < train_size:
                row.append(i)
            else:
                row.append(i + vocab_size)
            col.append(train_size + j)
            idf = log(1.0 * len(docs_of_words) / word_doc_freq[vocab[j]])
            weight_tfidf.append(freq * idf)
            doc_word_set.add(word)
    return weight_tfidf


def compute_weights_with_PMI(docs_of_words, rela_pair_count_str, word_window_freq, train_size, word_id_map, min_count1, max_count1, count_mean1, count_std1, row, col):
    weight = []
    errors = 0
    num_window = len(docs_of_words)
    for key, count in rela_pair_count_str.items():
        try:
            temp = key.split(',')
            i = temp[0]
            j = temp[1]

            if i in word_window_freq and j in word_window_freq:
                word_freq_i = word_window_freq[i]
                word_freq_j = word_window_freq[j]
                pmi = log(
                    1.0
                    * count
                    / num_window
                    / (1.0 * word_freq_i * word_freq_j / num_window ** 2)
                )
                if pmi <= 0:
                    continue

                row.append(train_size + word_id_map[i])
                col.append(train_size + word_id_map[j])
                if key in rela_pair_count_str:
                    wei = (rela_pair_count_str[key] -
                           min_count1) / (max_count1 - min_count1)
                    wei = (rela_pair_count_str[key]-count_mean1) / count_std1
                    weight.append(wei)
        except:
            errors += 1

    print(f'[INFO] Error in Compute Weights: {errors}')
    return weight



from wheel.bdist_wheel import bdist_wheel as _bdist_wheel
from setuptools import find_packages, setup
import platform
import pathlib


class MyBdistWheel(_bdist_wheel):
    def finalize_options(self):
        _bdist_wheel.finalize_options(self)
        self.root_is_pure = False


if __name__ == "__main__":
    os_tag = {
        "Windows": "win_amd64",
        "Darwin": "macosx_x86_64",
        "Linux": "manylinux1_x86_64"
    }

    # The directory containing this file
    # Reads the content of your README.md into a variable to be used in the setup below
    HERE = pathlib.Path(__file__).parent

    # The text of the README file
    README = (HERE / "README.md").read_text()

    install_requirement = [
        "gensim >= 4.1.2",
        "nltk >= 3.5",
        "scikit-learn >= 0.23.2",
        "stanfordcorenlp >= 3.9.1.1",
        "tabulate >= 0.8.9",
        "matplotlib >= 3.1.1",
        "torch >= 1.11.0",
    ]

    # This call to setup() does all the work
    setup(
        author="Vitor Meriat",
        author_email="vitormeriat@gmail.com",
        # Should match the package folder
        name="text4gcn",
        # Should match the package folder
        # Packages=["text4gcn"],
        packages=find_packages(
            ".",
            exclude=(
                "examples.*",
                "examples",
                "tests",
                "test.py",
            ),
        ),
        # Important for updates
        version="1.0.0",
        # Should match your chosen license
        license="MIT",
        description="A library for building text graphs for the application of Graph Neural Networks (GNN), in the context of text classification in natural language processing",
        # Loads your README.md
        long_description=README,
        # README.md is of type 'markdown'
        long_description_content_type="text/markdown",
        url="https://github.com/vitormeriat/text4gcn",
        keywords=["tgcn", "gcn", "gnn", "nlp",
                  "text_classification", "text_graph"],
        classifiers=[
            "License :: OSI Approved :: MIT License",
            #"Programming Language :: Python :: 3",
            "Programming Language :: Python :: 3.8",
            "Operating System :: OS Independent",
        ],
        include_package_data=True,
        install_requires=install_requirement,
        platforms=os_tag[platform.system()],
        # entry_points={
        #     "console_scripts": [
        #         "text4gcn=text4gcn.__main__:main",
        #     ]
        # },
    )
    print(
        "Text4GCN Python library installation finished. Please manually check Stanford CoreNLP"
        "(https://stanfordnlp.github.io/CoreNLP/) is installed and "
        "running in your environment."
    )



from typing import Any, Dict, Iterable, Optional
from collections import Counter
from nltk import word_tokenize
from typing import List

def flatten_nested_iterables(iterables_of_iterables: Iterable[Iterable[Any]]) -> Iterable[Any]:
    return [item for sublist in iterables_of_iterables for item in sublist]


from text4gcn.models.train_eval import TextGCNTrainer
from text4gcn.models.utils import LoadData
from text4gcn.models.layers import Layer
from text4gcn.models.model import GCN
import numpy as np
import torch as th
import argparse

# # parser = argparse.ArgumentParser(description='TextGCN')
# # #parser.add_argument('--dataset', type=str, default='citeseer', help='Dataset to train')
# # parser.add_argument('--dataset', type=str, help='Dataset to train')
# # parser.add_argument('--lr', type=float, default=0.01, help='Initial learing rate')
# # parser.add_argument('--epoches', type=int, default=200, help='Number of traing epoches')
# # parser.add_argument('--hidden_dim', type=list, default=200, help='Dimensions of hidden layers')
# # parser.add_argument('--dropout', type=float, default=0.5, help='Dropout rate (1 - keep  probability)')
# # parser.add_argument('--weight_decay', type=float, default=5e-4, help='Weight for l2 loss on embedding matrix')
# # parser.add_argument('--log_interval', type=int, default=10, help='Print iterval')
# # parser.add_argument('--log_dir', type=str, default='experiments', help='Train/val loss and accuracy logs and plots')
# # parser.add_argument('--checkpoint_interval', type=int, default=20, help='Checkpoint saved interval')
# # parser.add_argument('--checkpoint_dir', type=str, default='checkpoints', help='Directory to save checkpoints')
# # args = parser.parse_args()


class GNN():
    """
        The graph2seq model consists the following components: 1) node embedding 2) graph embedding 3) decoding. # noqa
        Since the full pipeline will consist all parameters, so we will add prefix to the original parameters
         in each component as follows (except the listed four parameters):
            1) emb_ + parameter_name (eg: ``emb_input_size``)
            2) gnn_ + parameter_name (eg: ``gnn_direction_option``)
            3) dec_ + parameter_name (eg: ``dec_max_decoder_step``)
        Considering neatness, we will only present the four hyper-parameters which don't meet regulations.

    Examples
    -------
        # Build a vocab model from scratch
        >>> "It is just a how-to-use example."
        >>> from graph4nlp.pytorch.modules.config import get_basic_args
        >>> opt = get_basic_args(graph_construction_name="node_emb", graph_embedding_name="gat", decoder_name="stdrnn")
        >>> graph2seq = Graph2Seq.from_args(opt=opt, vocab_model=vocab_model, device=torch.device("cuda:0"))
        >>> batch_graph = [GraphData() for _ in range(2)]
        >>> tgt_seq = torch.Tensor([[1, 2, 3], [4, 5, 6]])
        >>> seq_out, _, _ = graph2seq(batch_graph=batch_graph, tgt_seq=tgt_seq)
        >>> print(seq_out.shape) # [2, 6, 5] (assume the vocabulary size is 5 and max_decoder_step is 6)

    Attributes
    ----------
    dataset : str
        first name of the person
    path : str
        family name of the person
    log_dir : int
        age of the person
    seed: int, default=1234
        The seed for random function.

    Methods
    -------
    config():
        Prints the person's name and age.
    fit():
        Prints the person's name and age.
    """

    def __init__(
        self,
        dataset,
        path,
        log_dir,
        builder,
        lr=0.02,
        nhid=200,
        seed=1234,
        epoches=200,
        dropout=0.5,
        val_ratio=0.1,
        layer=Layer.GCN,
        early_stopping=10
    ) -> None:

        self.early_stopping = early_stopping
        self.val_ratio = val_ratio
        self.dataset = dataset
        self.epoches = epoches
        self.dropout = dropout
        self.log_dir = log_dir
        self.builder = builder
        self.layer = layer
        self.seed = seed
        self.nhid = nhid
        self.path = path
        self.lr = lr

    def config(self):
        print(f"\n{'='*30} MODEL CONFIGURATION\n")

        parser = argparse.ArgumentParser(description='TextGCN')
        self.args = parser.parse_args()

        self.args.early_stopping = self.early_stopping
        self.args.seed = np.random.randint(1, 100000)
        self.args.val_ratio = self.val_ratio
        self.args.max_epoch = self.epoches
        self.args.dataset = self.dataset
        self.args.dropout = self.dropout
        self.args.log_dir = self.log_dir
        self.args.layer = self.layer
        self.args.nhid = self.nhid
        self.args.lr = self.lr
        self.args.device = th.device(
            'cuda') if th.cuda.is_available() else th.device('cpu')

        print(f'{self.args}\n')

    def fit(self):
        self.config()

        predata = LoadData(
            path=self.path,
            dataset=self.dataset,
            builder=self.builder).load_corpus()

        if self.layer == Layer.GCN:
            model = GCN
        else:
            raise TypeError(
                "Invalide Layer. In this version only the GCN layer is valid!")

        framework = TextGCNTrainer(
            model=model,
            args=self.args,
            pre_data=predata
        )

        framework.fit()



from sklearn.feature_extraction.text import TfidfVectorizer
from typing import List, Iterable, Tuple
from collections import OrderedDict
import numpy as np
import pickle as pkl
from math import log


def word_to_vectors(vocabulary, word_vectors):
    return OrderedDict((word, vec.tolist()) for word, vec in zip(vocabulary, word_vectors))


def extract_vocabulary(docs_of_words: Iterable[List[str]]) -> List[str]:
    vocabulary = OrderedDict()
    for words in docs_of_words:
        vocabulary.update((word, None) for word in words)
    return list(vocabulary.keys())


def extract_tf_idf_word_vectors(word_definitions: List[str], max_features: int) -> List[np.ndarray]:
    tf_idf_vectorizer = TfidfVectorizer(max_features=max_features)
    return tf_idf_vectorizer.fit_transform(word_definitions).toarray()


def extract_word_definitions(vocabulary: List[str]) -> List[str]:
    from nltk.corpus import wordnet
    #from nltk import download
    #temporary_nltk_folder = 'venv/nltk_data/'
    #download(info_or_id='wordnet', download_dir=temporary_nltk_folder)
    #download(info_or_id='wordnet')
    #download(info_or_id='omw-1.4')

    merged_definitions_of_words = []
    for word in vocabulary:
        syn_sets_of_word = wordnet.synsets(word.strip())
        word_definitions = [syn_set.definition()
                            for syn_set in syn_sets_of_word]
        merged_definitions_of_word = ' '.join(word_definitions) if word_definitions else '<PAD>'

        merged_definitions_of_words.append(merged_definitions_of_word)
    # rmtree(temporary_nltk_folder)
    return merged_definitions_of_words


def build_word_doc_list(docs_of_words):
    word_doc_list = {}
    for i in range(len(docs_of_words)):
        words = docs_of_words[i]
        appeared = set()
        for word in words:
            if word in appeared:
                continue
            if word in word_doc_list:
                doc_list = word_doc_list[word]
                doc_list.append(i)
                word_doc_list[word] = doc_list
            else:
                word_doc_list[word] = [i]
            appeared.add(word)
    return word_doc_list

def define_word_doc_freq(docs_of_words):
    word_doc_list = build_word_doc_list(docs_of_words)
    return {
        word: len(doc_list) for word, doc_list in word_doc_list.items()
    }

def build_word_window_freq(docs_of_words):
    word_window_freq = {}
    for window in docs_of_words:
        appeared = set()
        for i in range(len(window)):
            if window[i] in appeared:
                continue
            if window[i] in word_window_freq:
                word_window_freq[window[i]] += 1
            else:
                word_window_freq[window[i]] = 1
            appeared.add(window[i])
    return word_window_freq

def relation_pair_statitcs(rela_pair_count_str):
    max_count1 = 0.0
    min_count1 = 0.0
    count1 = []
    for key, value in rela_pair_count_str.items():
        if rela_pair_count_str[key] > max_count1:
            max_count1 = rela_pair_count_str[key]
        if value < min_count1:
            min_count1 = rela_pair_count_str[key]
        count1.append(rela_pair_count_str[key])

    count_mean1 = np.mean(count1)
    count_std1 = np.std(count1, ddof=1)
    return min_count1, max_count1, count_mean1, count_std1

def get_weight_tfidf(docs_of_words, word_id_map, doc_word_freq, train_size, vocab_size, word_doc_freq, vocab, row, col):
    weight_tfidf = []
    for i in range(len(docs_of_words)):
        words = docs_of_words[i]
        doc_word_set = set()
        for word in words:
            if word in doc_word_set:
                continue
            j = word_id_map[word]
            key = f'{str(i)},{str(j)}'
            freq = doc_word_freq[key]
            if i < train_size:
                row.append(i)
            else:
                row.append(i + vocab_size)
            col.append(train_size + j)
            idf = log(1.0 * len(docs_of_words) / word_doc_freq[vocab[j]])
            weight_tfidf.append(freq * idf)
            doc_word_set.add(word)
    return weight_tfidf

def get_doc_word_freq(docs_of_words, word_id_map):
    doc_word_freq = {}
    for doc_id in range(len(docs_of_words)):
        words = docs_of_words[doc_id]
        for word in words:
            word_id = word_id_map[word]
            doc_word_str = f'{str(doc_id)},{str(word_id)}'
            if doc_word_str in doc_word_freq:
                doc_word_freq[doc_word_str] += 1
            else:
                doc_word_freq[doc_word_str] = 1
    return doc_word_freq